# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

import pytest
import torch
from lhotse import CutSet, MonoCut, SupervisionSegment
from lhotse.testing.dummies import DummyManifest, dummy_cut
from lhotse.testing.random import deterministic_rng
from omegaconf import DictConfig

from nemo.collections.asr.data.audio_to_text_lhotse_prompted import (
    PromptedAudioToTextLhotseDataset,
    PromptedAudioToTextMiniBatch,
)
from nemo.collections.asr.models.aed_multitask_models import EncDecMultiTaskModel
from nemo.collections.asr.parts.submodules import multitask_beam_decoding as beam_decode
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis
from nemo.collections.asr.parts.utils.streaming_utils import FrameBatchMultiTaskAED
from nemo.collections.asr.parts.utils.timestamp_utils import process_aed_timestamp_outputs
from nemo.collections.common.prompts.canary import CanaryPromptFormatter, canary
from nemo.collections.common.prompts.canary2 import Canary2PromptFormatter, canary2
from nemo.collections.common.tokenizers import CanaryTokenizer


@pytest.fixture()
def asr_model(test_data_dir):
    preprocessor = {
        'cls': 'nemo.collections.asr.modules.AudioToMelSpectrogramPreprocessor',
        'params': {"window_size": 0.02, "window_stride": 0.01, "features": 64},
    }

    model_defaults = {'asr_enc_hidden': 128, 'lm_enc_hidden': 64, 'lm_dec_hidden': 64}

    encoder = {
        'cls': 'nemo.collections.asr.modules.ConformerEncoder',
        'params': {
            'feat_in': 64,
            'n_layers': 1,
            'd_model': model_defaults['asr_enc_hidden'],
            'subsampling': 'dw_striding',
            'subsampling_factor': 2,
            'ff_expansion_factor': 4,
            'self_attention_model': 'rel_pos',
            'n_heads': 4,
            'conv_kernel_size': 9,
        },
    }

    transf_decoder = {
        '_target_': 'nemo.collections.asr.modules.transformer.get_nemo_transformer',
        'model_name': None,
        'pretrained': False,
        'encoder': None,
        'pre_ln_final_layer_norm': True,
        'config_dict': {
            'max_sequence_length': 512,
            'num_token_types': 0,
            'hidden_size': model_defaults['lm_dec_hidden'],
            'inner_size': 4 * model_defaults['lm_dec_hidden'],
            'num_layers': 1,
            'num_attention_heads': 2,
            'pre_ln': True,
            'vocab_size': None,
        },
    }

    head = {
        '_target_': 'nemo.collections.asr.parts.submodules.token_classifier.TokenClassifier',
        'num_layers': 1,
        'activation': 'relu',
        'log_softmax': True,
        'hidden_size': model_defaults['lm_dec_hidden'],
        'num_classes': None,
    }

    decoding = {'strategy': 'beam', 'beam': {'beam_size': 1}}

    # os.path.join(test_data_dir, "asr", "tokenizers", "an4_wpe_128")
    tokenizer = {
        'dir': None,
        'type': 'agg',
        'langs': {
            'spl_tokens': {
                'dir': os.path.join(test_data_dir, "asr", "tokenizers", "canary"),
                'type': 'bpe',
            },
            'en': {
                'dir': os.path.join(test_data_dir, "asr", "tokenizers", "an4_wpe_128"),
                'type': 'wpe',
            },
            'de': {
                'dir': os.path.join(test_data_dir, "asr", "tokenizers", "an4_wpe_128"),
                'type': 'wpe',
            },
        },
        'custom_tokenizer': {
            '_target_': 'nemo.collections.common.tokenizers.canary_tokenizer.CanaryTokenizer',
            'tokenizers': None,
        },
    }

    optim = {
        'name': 'adamw',
        'lr': 1e-4,
    }

    loss = {
        '_target_': 'nemo.collections.common.losses.smoothed_cross_entropy.SmoothedCrossEntropyLoss',
        'label_smoothing': 0.0,
    }

    modelConfig = DictConfig(
        {
            'prompt_format': 'canary',
            'prompt_defaults': [
                {"role": "user", "slots": {"source_lang": "en", "target_lang": "en", "task": "asr", "pnc": "yes"}}
            ],
            'sample_rate': 16000,
            'preprocessor': DictConfig(preprocessor),
            'model_defaults': DictConfig(model_defaults),
            'encoder': DictConfig(encoder),
            'transf_decoder': DictConfig(transf_decoder),
            'head': DictConfig(head),
            'tokenizer': DictConfig(tokenizer),
            'decoding': DictConfig(decoding),
            'optim': DictConfig(optim),
            'loss': DictConfig(loss),
        }
    )

    model_instance = EncDecMultiTaskModel(cfg=modelConfig)
    model_instance.configure_optimizers()
    return model_instance


class TestEncDecMultiTaskModel:
    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_constructor(self, asr_model):
        asr_model.train()
        # Check to/from config_dict:
        confdict = asr_model.to_config_dict()
        instance2 = EncDecMultiTaskModel.from_config_dict(confdict)
        assert isinstance(instance2, EncDecMultiTaskModel)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_forward(self, asr_model):
        torch.manual_seed(0)
        asr_model = asr_model.eval()

        asr_model.preprocessor.featurizer.dither = 0.0
        asr_model.preprocessor.featurizer.pad_to = 0

        asr_model.compute_eval_loss = False

        input_signal = torch.randn(size=(4, 512))
        length = torch.randint(low=321, high=500, size=[4])

        targets = torch.randint(low=0, high=100, size=[4, 10])
        targets_len = torch.randint(low=1, high=10, size=[4])

        with torch.no_grad():
            # batch size 1
            logprobs_instance = []
            for i in range(input_signal.size(0)):
                log_probs, _, _, _ = asr_model.forward(
                    input_signal=input_signal[i : i + 1],
                    input_signal_length=length[i : i + 1],
                    transcript=targets[i : i + 1],
                    transcript_length=targets_len[i : i + 1],
                )
                logprobs_instance.append(log_probs)
            logits_instance = torch.cat(logprobs_instance, 0)

            # batch size 4
            logprobs_batch, _, _, _ = asr_model.forward(
                input_signal=input_signal,
                input_signal_length=length,
                transcript=targets,
                transcript_length=targets_len,
            )

        assert logits_instance.shape == logprobs_batch.shape
        diff = torch.mean(torch.abs(logits_instance - logprobs_batch))
        assert diff <= 1e-5
        diff = torch.max(torch.abs(logits_instance - logprobs_batch))
        assert diff <= 1e-5

    @pytest.mark.unit
    def test_training_step(self, deterministic_rng, asr_model):
        cuts = CutSet(
            [
                dummy_cut(
                    0,
                    duration=1.0,
                    with_data=True,
                    supervisions=[
                        SupervisionSegment(
                            id="cut-0", recording_id="cut-0", start=0, duration=1.0, text="short", language="en"
                        )
                    ],
                ),
                dummy_cut(
                    1,
                    duration=5.0,
                    recording_duration=5.0,
                    with_data=True,
                    supervisions=[
                        SupervisionSegment(
                            id="cut-1",
                            recording_id="cut-1",
                            start=0,
                            duration=5.0,
                            text="a very long transcript",
                            language="en",
                        )
                    ],
                ),
            ]
        )
        for c in cuts:
            c.source_lang = "en"
            c.target_lang = "en"
            c.task = "asr"
            c.pnc = "no"
        dataset = PromptedAudioToTextLhotseDataset(
            tokenizer=asr_model.tokenizer, prompt=CanaryPromptFormatter(asr_model.tokenizer)
        )
        batch = dataset[cuts]

        ans = asr_model.training_step(batch, batch_nb=0)
        assert list(ans.keys()) == ["loss"]
        assert torch.is_tensor(ans["loss"])

    @pytest.mark.unit
    def test_validation_step(self, deterministic_rng, asr_model):
        cuts = CutSet(
            [
                dummy_cut(
                    0,
                    duration=1.0,
                    with_data=True,
                    supervisions=[
                        SupervisionSegment(
                            id="cut-0", recording_id="cut-0", start=0, duration=1.0, text="short", language="en"
                        )
                    ],
                ),
                dummy_cut(
                    1,
                    duration=5.0,
                    recording_duration=5.0,
                    with_data=True,
                    supervisions=[
                        SupervisionSegment(
                            id="cut-1",
                            recording_id="cut-1",
                            start=0,
                            duration=5.0,
                            text="a very long transcript",
                            language="en",
                        )
                    ],
                ),
            ]
        )
        for c in cuts:
            c.source_lang = "en"
            c.target_lang = "en"
            c.task = "asr"
            c.pnc = "no"
        dataset = PromptedAudioToTextLhotseDataset(
            tokenizer=asr_model.tokenizer, prompt=CanaryPromptFormatter(asr_model.tokenizer)
        )
        batch = dataset[cuts]

        with torch.no_grad():
            ans = asr_model.validation_pass(batch, batch_idx=0)
        print(ans)
        assert set(ans.keys()) == set(
            [
                "val_loss",
                "val_wer",
                "val_wer_num",
                "val_wer_denom",
                "val_bleu",
                "val_bleu_pred_len",
                "val_bleu_target_len",
                "val_bleu_num",
                "val_bleu_denom",
            ]
        )

    @pytest.mark.unit
    def test_save_restore_artifact(self, asr_model):
        asr_model.train()

        with tempfile.TemporaryDirectory() as tmp_dir:
            path = os.path.join(tmp_dir, 'aed_bpe.nemo')
            asr_model.save_to(path)

            new_model = EncDecMultiTaskModel.restore_from(path)
            assert isinstance(new_model, type(asr_model))

            assert len(new_model.tokenizer.tokenizer.get_vocab()) == 32 + 128 + 128

    @pytest.mark.unit
    def test_restore_with_timestamps_asr_model(self, canary_1b_v2):
        assert canary_1b_v2.timestamps_asr_model is not None

    # @pytest.mark.with_downloads()
    # @pytest.mark.unit
    # def test_save_restore_artifact_change_vocab(self, asr_model, test_data_dir):
    #     asr_model.train()
    #
    #     with tempfile.TemporaryDirectory() as tmpdir:
    #         tokenizer_dir = os.path.join(test_data_dir, "asr", "tokenizers", "an4_spe_128")
    #         asr_model.change_vocabulary(new_tokenizer_dir=tokenizer_dir, new_tokenizer_type='bpe')
    #
    #         save_path = os.path.join(tmpdir, 'ctc_bpe.nemo')
    #         asr_model.train()
    #         asr_model.save_to(save_path)
    #
    #         new_model = EncDecMultiTaskModel.restore_from(save_path)
    #         assert isinstance(new_model, type(asr_model))
    #         assert isinstance(new_model.tokenizer, tokenizers.SentencePieceTokenizer)
    #         assert new_model.model_path.endswith('_tokenizer.model')
    #         assert new_model.vocab_path.endswith('_vocab.txt')
    #         assert new_model.spe_vocab_path.endswith('_tokenizer.vocab')

    # @pytest.mark.with_downloads()
    # @pytest.mark.unit
    # def test_save_restore_artifact_agg(self, asr_model, test_data_dir):
    #     tokenizer_dir = os.path.join(test_data_dir, "asr", "tokenizers", "an4_spe_128")
    #     tok_en = {"dir": tokenizer_dir, "type": "wpe"}
    #     # the below is really an english tokenizer but we pretend it is spanish
    #     tok_es = {"dir": tokenizer_dir, "type": "wpe"}
    #     tcfg = DictConfig({"type": "agg", "langs": {"en": tok_en, "es": tok_es}})
    #     with tempfile.TemporaryDirectory() as tmpdir:
    #         asr_model.change_vocabulary(new_tokenizer_dir=tcfg, new_tokenizer_type="agg")
    #
    #         save_path = os.path.join(tmpdir, "ctc_agg.nemo")
    #         asr_model.train()
    #         asr_model.save_to(save_path)
    #
    #         new_model = EncDecMultiTaskModel.restore_from(save_path)
    #         assert isinstance(new_model, type(asr_model))
    #         assert isinstance(new_model.tokenizer, tokenizers.AggregateTokenizer)
    #
    #         # should be double
    #         assert new_model.tokenizer.tokenizer.vocab_size == 254
    #         assert len(new_model.tokenizer.tokenizer.get_vocab()) == 254

    # @pytest.mark.with_downloads()
    # @pytest.mark.unit
    # def test_vocab_change(self, test_data_dir, asr_model):
    #     with tempfile.TemporaryDirectory() as tmpdir:
    #         old_tokenizer_dir = os.path.join(test_data_dir, "asr", "tokenizers", "an4_wpe_128", 'vocab.txt')
    #         new_tokenizer_dir = os.path.join(tmpdir, 'tokenizer')
    #
    #         os.makedirs(new_tokenizer_dir, exist_ok=True)
    #         shutil.copy2(old_tokenizer_dir, new_tokenizer_dir)
    #
    #         nw1 = asr_model.num_weights
    #         asr_model.change_vocabulary(new_tokenizer_dir=new_tokenizer_dir, new_tokenizer_type='wpe')
    #         # No change
    #         assert nw1 == asr_model.num_weights
    #
    #         with open(os.path.join(new_tokenizer_dir, 'vocab.txt'), 'a+') as f:
    #             f.write("!\n")
    #             f.write('$\n')
    #             f.write('@\n')
    #
    #         asr_model.change_vocabulary(new_tokenizer_dir=new_tokenizer_dir, new_tokenizer_type='wpe')
    #
    #         # rnn embedding + joint + bias
    #         pred_embedding = 3 * (asr_model.decoder.pred_hidden)
    #         joint_joint = 3 * (asr_model.joint.joint_hidden + 1)
    #         assert asr_model.num_weights == (nw1 + (pred_embedding + joint_joint))

    @pytest.mark.unit
    def test_decoding_change(self, asr_model):
        assert isinstance(asr_model.decoding.decoding, beam_decode.TransformerAEDBeamInfer)

        new_strategy = DictConfig({})
        new_strategy.strategy = 'beam'
        new_strategy.beam = DictConfig({'beam_size': 5})
        asr_model.change_decoding_strategy(decoding_cfg=new_strategy)
        assert isinstance(asr_model.decoding.decoding, beam_decode.TransformerAEDBeamInfer)
        assert asr_model.decoding.decoding.search_type == "default"

    @pytest.mark.unit
    def test_prompt_change(self, asr_model):
        assert asr_model.prompt_format == 'canary'
        assert isinstance(asr_model.prompt, CanaryPromptFormatter)

        # Default change prompt
        asr_model.change_prompt()
        assert asr_model.cfg.prompt_defaults is None

        prompt_defaults = asr_model.prompt.get_default_dialog_slots()
        prompt_defaults[0]['slots']['pnc'] = 'no'
        asr_model.change_prompt(prompt_defaults=prompt_defaults)

        assert asr_model.cfg.prompt_defaults[0]['slots']['pnc'] == 'no'

    @pytest.mark.unit
    def test_prompt_change_subclass(self, asr_model):
        assert asr_model.prompt_format == 'canary'
        assert isinstance(asr_model.prompt, CanaryPromptFormatter)

        class CanaryPromptFormatterSubclass(CanaryPromptFormatter):
            NAME = "canary-unit-test-stub-format"

        # Default change prompt
        asr_model.change_prompt()
        assert asr_model.cfg.prompt_defaults is None

        prompt_defaults = asr_model.prompt.get_default_dialog_slots()
        prompt_defaults[0]['slots']['pnc'] = 'no'
        asr_model.change_prompt(prompt_format='canary-unit-test-stub-format', prompt_defaults=prompt_defaults)

        assert asr_model.cfg.prompt_format == 'canary-unit-test-stub-format'
        assert asr_model.cfg.prompt_defaults[0]['slots']['pnc'] == 'no'
        assert isinstance(asr_model.prompt, CanaryPromptFormatterSubclass)

        user_prompt = asr_model.prompt.get_default_dialog_slots()[0]
        slots = user_prompt['slots']
        slots['source_lang'] = 'en'
        slots['target_lang'] = 'en'
        slots['task'] = 'asr'
        slots['pnc'] = 'no'
        ans = asr_model.prompt.encode_dialog([user_prompt])
        recovered = asr_model.tokenizer.ids_to_text(ans["input_ids"])
        assert recovered == "<|startoftranscript|><|en|><|transcribe|><|en|><|nopnc|>"

    @pytest.mark.unit
    def test_transcribe_single_file(self, asr_model, test_data_dir):
        audio_file = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an46-mmap-b.wav")

        # Numpy array test
        outputs = asr_model.transcribe(audio_file, batch_size=1)
        assert len(outputs) == 1
        assert isinstance(outputs[0].text, str)

    @pytest.mark.unit
    def test_transcribe_single_file_translation(self, asr_model, test_data_dir):
        audio_file = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an46-mmap-b.wav")

        # Numpy array test
        outputs = asr_model.transcribe(audio_file, batch_size=1, task="ast", source_lang='en', target_lang='de')
        assert len(outputs) == 1
        assert isinstance(outputs[0].text, str)

    @pytest.mark.unit
    def test_transcribe_return_hypothesis(self, asr_model, test_data_dir):
        audio_file = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an46-mmap-b.wav")

        # Numpy array test
        outputs = asr_model.transcribe(audio_file, batch_size=1, return_hypotheses=True)
        assert len(outputs) == 1
        assert isinstance(outputs[0], Hypothesis)

        hyp = outputs[0]
        assert isinstance(hyp.text, str)
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert hyp.alignments is None

    # test transcribe with canary2 model with torch tensor input
    @pytest.mark.unit
    def test_transcribe_with_canary2_model_with_torch_tensor_input(self, canary_1b_v2, test_data_dir):
        import soundfile as sf

        audio_file = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an46-mmap-b.wav")
        audio, sr = sf.read(audio_file, dtype='float32')

        # Numpy array test
        outputs = canary_1b_v2.transcribe(audio, batch_size=1)
        assert len(outputs) == 1
        assert isinstance(outputs[0].text, str)

    @pytest.mark.unit
    def test_transcribe_tensor(self, asr_model, test_data_dir):
        # Load audio file
        import soundfile as sf

        audio_file = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an46-mmap-b.wav")
        audio, sr = sf.read(audio_file, dtype='float32')

        # Numpy array test
        outputs = asr_model.transcribe(audio, batch_size=1)
        assert len(outputs) == 1
        assert isinstance(outputs[0].text, str)

    @pytest.mark.unit
    def test_build_tokenizer(self, asr_model, test_data_dir):
        # Load audio file
        task_tokens = ["ast", "asr"]
        lang_tokens = ["en", "es", "de", "fr"]
        tokens = task_tokens + lang_tokens
        spl_tokenizer_from_build = CanaryTokenizer.build_special_tokenizer(tokens, test_data_dir)

        tokenizer_cfg = {'dir': os.path.join(test_data_dir), 'type': 'bpe'}
        spl_tokenizer_from_load = asr_model._make_tokenizer(tokenizer_cfg, "spl_tokens")[0]

        tokens += ["<|nospeech|>", "<pad>", "<|endoftext|>", "<|startoftranscript|>", "<|pnc|>", "<|nopnc|>"]

        ids1 = [spl_tokenizer_from_build.tokens_to_ids(t)[0] for t in tokens]
        ids2 = [spl_tokenizer_from_load.tokens_to_ids(t)[0] for t in tokens]

        for i, j in zip(ids1, ids2):
            assert i == j

    @pytest.mark.unit
    def test_predict_step(self, asr_model, test_data_dir):
        cuts = DummyManifest(CutSet, begin_id=0, end_id=1, with_data=True)
        c = cuts[0]
        c.supervisions[0].language = "en"
        c.source_lang = "en"
        c.target_lang = "en"
        c.task = "asr"
        c.pnc = "no"
        dataset = PromptedAudioToTextLhotseDataset(
            tokenizer=asr_model.tokenizer, prompt=CanaryPromptFormatter(asr_model.tokenizer)
        )
        batch = dataset[cuts]

        # Numpy array test
        outputs = asr_model.predict_step(batch)
        print(outputs)
        assert len(outputs) == 1
        assert len(outputs[0]) == 2
        assert isinstance(outputs[0][0], MonoCut)
        assert isinstance(outputs[0][1].text, str)

    @pytest.mark.unit
    def test_FrameBatchMultiTaskAED(self, asr_model, test_data_dir):
        model = FrameBatchMultiTaskAED(asr_model, batch_size=1)

        audio_file = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an46-mmap-b.wav")
        meta = {
            'audio_filepath': audio_file,
            'duration': 100000,
            'source_lang': 'en',
            'taskname': 'asr',
            'target_lang': 'en',
            'pnc': 'yes',
            'answer': 'nothing',
        }
        model.read_audio_file(audio_file, delay=0.0, model_stride_in_secs=40.0, meta_data=meta)
        outputs = model.transcribe()
        assert isinstance(outputs, Hypothesis)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_FrameBatchMultiTaskAED_with_timestamps(self, canary_1b_flash):
        canary_1b_flash.eval()
        model = FrameBatchMultiTaskAED(
            canary_1b_flash,
            frame_len=10.0,
            total_buffer=10.0,
            batch_size=8,
        )

        audio_file = "/home/TestData/asr/longform/earnings22/sample_4469669.wav"

        meta = {
            'audio_filepath': audio_file,
            'duration': 100000,
            'source_lang': 'en',
            'taskname': 'asr',
            'target_lang': 'en',
            'pnc': 'yes',
            'answer': 'nothing',
            'timestamp': 'yes',
        }
        model_stride_in_secs = 0.01 * 8  # feature_stride in sec * model_stride
        model.read_audio_file(audio_file, delay=0.0, model_stride_in_secs=model_stride_in_secs, meta_data=meta)
        outputs = model.transcribe(timestamps=True)

        # check hypothesis object
        assert isinstance(outputs, Hypothesis)

        # check part of transcript
        assert outputs.text[:13] == "Now it's time", f"{outputs}"

        # check timestamps
        assert outputs.timestamp['segment'][0]['start'] == pytest.approx(5.68)
        assert outputs.timestamp['segment'][0]['end'] == pytest.approx(9.68)


@pytest.mark.unit
def test_prompted_dataset(asr_model):
    dataset = PromptedAudioToTextLhotseDataset(
        tokenizer=asr_model.tokenizer, prompt=CanaryPromptFormatter(asr_model.tokenizer)
    )

    cuts = DummyManifest(CutSet, begin_id=0, end_id=3, with_data=True)

    c = cuts[0]
    c.supervisions[0].language = "en"
    c.source_lang = "en"
    c.target_lang = "en"
    c.task = "asr"
    c.pnc = "no"

    c = cuts[1]
    c.supervisions[0].language = "de"
    c.supervisions[0].text = "unerheblich"
    c.source_lang = "en"
    c.target_lang = "de"
    c.taskname = "ast"  # note: testing for "taskname" as we support it together with "task"
    c.pnc = "yes"

    c = cuts[2]
    c.supervisions[0].language = "en"
    c.supervisions[0].text = ""
    c.source_lang = "en"
    c.target_lang = "en"
    c.task = "asr"
    c.pnc = "yes"

    batch = dataset[cuts]

    assert isinstance(batch, PromptedAudioToTextMiniBatch)
    assert batch.audio.shape == (3, 16000)
    assert batch.audio_lens.tolist() == [16000, 16000, 16000]

    # Test example 0 (transcription)
    i = 0
    assert (
        asr_model.tokenizer.ids_to_text(batch.prompt[i]) == '<|startoftranscript|><|en|><|transcribe|><|en|><|nopnc|>'
    )
    assert batch.prompt_lens[i] == 5
    assert asr_model.tokenizer.ids_to_text(batch.transcript[i]) == 'i##r##r##el##e##v##a##nt<pad><pad>'
    assert batch.transcript_lens[i] == 8
    assert (
        asr_model.tokenizer.ids_to_text(batch.prompted_transcript[i])
        == '<|startoftranscript|><|en|><|transcribe|><|en|><|nopnc|>i##r##r##el##e##v##a##nt<|endoftext|><pad><pad>'
    )
    assert batch.prompted_transcript_lens[i] == 14

    # Test example 1 (translation)
    i = 1
    assert asr_model.tokenizer.ids_to_text(batch.prompt[i]) == '<|startoftranscript|><|en|><|translate|><|de|><|pnc|>'
    assert batch.prompt_lens[i] == 5
    assert asr_model.tokenizer.ids_to_text(batch.transcript[i]) == 'u##ne##r##h##e##b##l##i##c##h'
    assert batch.transcript_lens[i] == 10
    assert (
        asr_model.tokenizer.ids_to_text(batch.prompted_transcript[i])
        == '<|startoftranscript|><|en|><|translate|><|de|><|pnc|>u##ne##r##h##e##b##l##i##c##h<|endoftext|>'
    )
    assert batch.prompted_transcript_lens[i] == 16

    # Test example 2 (no transcript, e.g. noise)
    i = 2
    assert asr_model.tokenizer.ids_to_text(batch.prompt[i]) == '<|startoftranscript|><|en|><|transcribe|><|en|><|pnc|>'
    assert batch.prompt_lens[i] == 5
    assert asr_model.tokenizer.ids_to_text(batch.transcript[i]) == '<pad>' * 10
    assert batch.transcript_lens[i] == 0
    assert (
        asr_model.tokenizer.ids_to_text(batch.prompted_transcript[i])
        == '<|startoftranscript|><|en|><|transcribe|><|en|><|pnc|><|endoftext|>' + '<pad>' * 10
    )
    assert batch.prompted_transcript_lens[i] == 6


@pytest.fixture()
def canary2_tokenizer(asr_model, tmp_path):
    return CanaryTokenizer(
        {
            "spl_tokens": CanaryTokenizer.build_special_tokenizer(
                [
                    "<|startofcontext|>",
                    "<|en|>",
                    "<|de|>",
                    "<|pnc|>",
                    "<|nopnc|>",
                    "<|itn|>",
                    "<|noitn|>",
                    "<|diarize|>",
                    "<|nodiarize|>",
                    "<|timestamp|>",
                    "<|notimestamp|>",
                    "<|emo:undefined|>",
                    "<|emo:happy|>",
                ]
                # Timestamp frame special tokens
                + [f"<|{i}|>" for i in range(900)],
                tmp_path,
                force_rebuild=False,
            ),
            "en": asr_model.tokenizer.tokenizers_dict["en"],
            "de": asr_model.tokenizer.tokenizers_dict["de"],
        }
    )


@pytest.mark.unit
def test_prompted_dataset_canary2(canary2_tokenizer):
    dataset = PromptedAudioToTextLhotseDataset(
        tokenizer=canary2_tokenizer, prompt=Canary2PromptFormatter(canary2_tokenizer)
    )

    cuts = DummyManifest(CutSet, begin_id=0, end_id=4, with_data=True)

    # backward compatibility
    c = cuts[0]
    c.supervisions[0].language = "en"
    c.source_lang = "en"
    c.target_lang = "en"

    # new format
    c = cuts[1]
    c.supervisions[0].language = "en"
    c.supervisions[0].text = "asd"
    c.source_lang = "en"
    c.target_lang = "en"
    c.pnc = "yes"
    c.itn = "yes"
    c.diarize = "yes"
    c.timestamp = "yes"
    c.emotion = "<|emo:happy|>"
    c.decodercontext = ""

    # new format with extra context
    c = cuts[2]
    c.supervisions[0].language = "en"
    c.supervisions[0].text = "asd"
    c.source_lang = "en"
    c.target_lang = "en"
    c.pnc = "<|pnc|>"
    c.itn = "<|noitn|>"
    c.diarize = "<|diarize|>"
    c.timestamp = "<|timestamp|>"
    c.emotion = "<|emo:happy|>"
    c.decodercontext = "some decoder context"

    # transcript with timestamps
    c = cuts[3]
    c.supervisions[0].language = "en"
    c.supervisions[0].text = "<|0|> hello <|3|> <|4|> world <|5|>"
    c.source_lang = "en"
    c.target_lang = "en"
    c.pnc = "<|pnc|>"
    c.itn = "<|noitn|>"
    c.diarize = "<|diarize|>"
    c.timestamp = "<|timestamp|>"
    c.emotion = "<|emo:happy|>"
    c.decodercontext = "some decoder context"

    batch = dataset[cuts]

    assert isinstance(batch, PromptedAudioToTextMiniBatch)
    assert batch.audio.shape == (4, 16000)
    assert batch.audio_lens.tolist() == [16000, 16000, 16000, 16000]

    # Test example 0
    i = 0
    assert (
        canary2_tokenizer.ids_to_text(batch.prompt[i])
        == '<|startofcontext|><|startoftranscript|><|emo:undefined|><|en|><|en|><|pnc|><|noitn|><|notimestamp|><|nodiarize|><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad>'
    )
    assert batch.prompt_lens[i] == 9
    assert canary2_tokenizer.ids_to_text(batch.transcript[i]) == 'i##r##r##el##e##v##a##nt<pad><pad><pad><pad><pad>'
    assert batch.transcript_lens[i] == 8
    assert (
        canary2_tokenizer.ids_to_text(batch.prompted_transcript[i])
        == '<|startofcontext|><|startoftranscript|><|emo:undefined|><|en|><|en|><|pnc|><|noitn|><|notimestamp|><|nodiarize|>i##r##r##el##e##v##a##nt<|endoftext|><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad>'
    )
    assert batch.prompted_transcript_lens[i] == 18

    # Test example 1
    i = 1
    assert (
        canary2_tokenizer.ids_to_text(batch.prompt[i])
        == '<|startofcontext|><|startoftranscript|><|emo:happy|><|en|><|en|><|pnc|><|itn|><|timestamp|><|diarize|><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad>'
    )
    assert batch.prompt_lens[i] == 9
    assert (
        canary2_tokenizer.ids_to_text(batch.transcript[i])
        == 'a##s##d<pad><pad><pad><pad><pad><pad><pad><pad><pad><pad>'
    )
    assert batch.transcript_lens[i] == 3
    assert (
        canary2_tokenizer.ids_to_text(batch.prompted_transcript[i])
        == '<|startofcontext|><|startoftranscript|><|emo:happy|><|en|><|en|><|pnc|><|itn|><|timestamp|><|diarize|>a##s##d<|endoftext|><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad>'
    )
    assert batch.prompted_transcript_lens[i] == 13

    # Test example 2
    i = 2
    assert (
        canary2_tokenizer.ids_to_text(batch.prompt[i])
        == '<|startofcontext|>s##o##m##ed##e##c##o##d##erc##o##nt##e##x##t<|startoftranscript|><|emo:happy|><|en|><|en|><|pnc|><|noitn|><|timestamp|><|diarize|>'
    )
    assert batch.prompt_lens[i] == 25
    assert (
        canary2_tokenizer.ids_to_text(batch.transcript[i])
        == 'a##s##d<pad><pad><pad><pad><pad><pad><pad><pad><pad><pad>'
    )
    assert batch.transcript_lens[i] == 3
    assert (
        canary2_tokenizer.ids_to_text(batch.prompted_transcript[i])
        == '<|startofcontext|>s##o##m##ed##e##c##o##d##erc##o##nt##e##x##t<|startoftranscript|><|emo:happy|><|en|><|en|><|pnc|><|noitn|><|timestamp|><|diarize|>a##s##d<|endoftext|><pad><pad><pad><pad><pad><pad><pad><pad><pad><pad>'
    )
    assert batch.prompted_transcript_lens[i] == 29

    # Test example 3
    i = 3
    assert (
        canary2_tokenizer.ids_to_text(batch.prompt[i])
        == '<|startofcontext|>s##o##m##ed##e##c##o##d##erc##o##nt##e##x##t<|startoftranscript|><|emo:happy|><|en|><|en|><|pnc|><|noitn|><|timestamp|><|diarize|>'
    )
    assert batch.prompt_lens[i] == 25
    assert canary2_tokenizer.ids_to_text(batch.transcript[i]) == '<|0|>h##el##l##o<|3|><|4|>w##o##r##l##d<|5|>'
    assert batch.transcript_lens[i] == 13
    assert (
        canary2_tokenizer.ids_to_text(batch.prompted_transcript[i])
        == '<|startofcontext|>s##o##m##ed##e##c##o##d##erc##o##nt##e##x##t<|startoftranscript|><|emo:happy|><|en|><|en|><|pnc|><|noitn|><|timestamp|><|diarize|><|0|>h##el##l##o<|3|><|4|>w##o##r##l##d<|5|><|endoftext|>'
    )
    assert batch.prompted_transcript_lens[i] == 39


@pytest.mark.unit
def test_aed_timestamp_processing():
    # Create test hypothesis with timestamps
    hyp = Hypothesis(
        text="<|10|>hello<|15|> <|20|>world<|25|>",
        y_sequence=None,
        score=None,
        alignments=None,
        length=None,
        timestamp={},
    )

    # Process timestamps with default parameters
    processed = process_aed_timestamp_outputs(hyp)
    assert isinstance(processed, list)
    assert len(processed) == 1
    assert processed[0].text == "hello world"

    # Check word-level timestamps
    word_timestamps = processed[0].timestamp['word']
    assert len(word_timestamps) == 2

    # Check first word "hello"
    assert word_timestamps[0]['word'] == 'hello'
    assert word_timestamps[0]['start_offset'] == 10
    assert word_timestamps[0]['end_offset'] == 15
    assert word_timestamps[0]['start'] == 0.1  # 10 * 0.01
    assert word_timestamps[0]['end'] == 0.15  # 15 * 0.01

    # Check second word "world"
    assert word_timestamps[1]['word'] == 'world'
    assert word_timestamps[1]['start_offset'] == 20
    assert word_timestamps[1]['end_offset'] == 25
    assert word_timestamps[1]['start'] == 0.2  # 20 * 0.01
    assert word_timestamps[1]['end'] == 0.25  # 25 * 0.01

    # Check segment-level timestamps
    segments = processed[0].timestamp['segment']
    assert len(segments) == 1
    assert segments[0]['start_offset'] == 10
    assert segments[0]['end_offset'] == 25
    assert segments[0]['start'] == 0.1
    assert segments[0]['end'] == 0.25

    # Test with different window_stride and subsampling_factor
    hyp = Hypothesis(
        text="<|10|>hello<|15|> <|20|>world<|25|>",
        y_sequence=None,
        score=None,
        alignments=None,
        length=None,
        timestamp={},
    )
    processed = process_aed_timestamp_outputs(hyp, subsampling_factor=2, window_stride=0.02)
    word_timestamps = processed[0].timestamp['word']

    # Check timing calculations with new parameters
    assert word_timestamps[0]['start'] == 0.4  # 10 * 0.02 * 2
    assert word_timestamps[0]['end'] == 0.6  # 15 * 0.02 * 2
    assert word_timestamps[1]['start'] == 0.8  # 20 * 0.02 * 2
    assert word_timestamps[1]['end'] == 1.0  # 25 * 0.02 * 2

    # Test case when text doesn't contain timestamps
    hyp = Hypothesis(text="hello world", y_sequence=None, score=None, alignments=None, length=None, timestamp={})

    # Process timestamps with default parameters
    processed = process_aed_timestamp_outputs(hyp)
    assert isinstance(processed, list)
    assert len(processed) == 1
    assert processed[0].text == "hello world"

    # Verify no timestamps were extracted
    assert processed[0].timestamp['word'] == []
    assert processed[0].timestamp['segment'] == []


@pytest.mark.unit
def test_aed_forced_aligned_timestamps(canary_1b_v2):

    audio_file = "/home/TestData/asr/canary/dev-other-wav/8173-294714-0040.wav"
    audio_batch = [audio_file, audio_file]

    # Testing with batch_size=2 to avoid dynamic_chunking and test pure timestamps extraction
    # Dynamic chunking with timestamps are tested in other tests

    hypotheses = canary_1b_v2.transcribe(audio_batch, timestamps=False, batch_size=2)
    assert len(hypotheses) == 2
    assert hypotheses[0].timestamp == []
    assert hypotheses[1].timestamp == []

    ts_hypotheses = canary_1b_v2.transcribe(audio_batch, timestamps=True, batch_size=2)
    assert len(ts_hypotheses) == 2

    assert "word" in ts_hypotheses[0].timestamp
    assert "segment" in ts_hypotheses[0].timestamp
    assert "char" not in ts_hypotheses[0].timestamp

    assert ts_hypotheses[0].text == hypotheses[0].text

    assert len(ts_hypotheses[0].timestamp['word']) == len(ts_hypotheses[0].text.split())

    segment_count = 0
    segment_separators = ['.', '?', '!', '...']
    for sep in segment_separators:
        if sep in ts_hypotheses[0].text:
            segment_count += 1
    if ts_hypotheses[0].text.strip()[-1] not in segment_separators:
        segment_count += 1

    assert len(ts_hypotheses[0].timestamp['segment']) == segment_count
    assert [word_offset['word'] for word_offset in ts_hypotheses[0].timestamp['word']] == ts_hypotheses[0].text.split()
    assert " ".join([word_offset['word'] for word_offset in ts_hypotheses[0].timestamp['word']]) == " ".join(
        [segment_offset['segment'] for segment_offset in ts_hypotheses[0].timestamp['segment']]
    )

    assert ts_hypotheses[0].timestamp['segment'][0]['start'] == ts_hypotheses[0].timestamp['word'][0]['start']
    assert ts_hypotheses[0].timestamp['segment'][-1]['end'] == ts_hypotheses[0].timestamp['word'][-1]['end']
    assert (
        ts_hypotheses[0].timestamp['segment'][0]['start_offset']
        == ts_hypotheses[0].timestamp['word'][0]['start_offset']
    )
    assert (
        ts_hypotheses[0].timestamp['segment'][-1]['end_offset'] == ts_hypotheses[0].timestamp['word'][-1]['end_offset']
    )


@pytest.mark.unit
def test_aed_forced_aligned_timestamps_audio_tensor(canary_1b_v2):
    import librosa

    audio_file = "/home/TestData/asr/canary/dev-other-wav/8173-294714-0040.wav"

    audio_batch = [
        torch.from_numpy(librosa.load(audio_file, sr=16000)[0]),
        torch.from_numpy(librosa.load(audio_file, sr=16000)[0]),
    ]

    # Testing with batch_size=2 to avoid dynamic_chunking and test pure timestamps extraction
    # Dynamic chunking with timestamps are tested in other tests

    hypotheses = canary_1b_v2.transcribe(audio_batch, timestamps=False, batch_size=2)
    assert len(hypotheses) == 2
    assert hypotheses[0].timestamp == []
    assert hypotheses[1].timestamp == []

    ts_hypotheses = canary_1b_v2.transcribe(audio_batch, timestamps=True, batch_size=2, enable_chunking=False)
    assert len(ts_hypotheses) == 2

    assert "word" in ts_hypotheses[0].timestamp
    assert "segment" in ts_hypotheses[0].timestamp
    assert "char" not in ts_hypotheses[0].timestamp

    assert ts_hypotheses[0].text == hypotheses[0].text

    assert len(ts_hypotheses[0].timestamp['word']) == len(ts_hypotheses[0].text.split())

    segment_count = 0
    segment_separators = ['.', '?', '!', '...']
    for sep in segment_separators:
        if sep in ts_hypotheses[0].text:
            segment_count += 1
    if ts_hypotheses[0].text.strip()[-1] not in segment_separators:
        segment_count += 1

    assert len(ts_hypotheses[0].timestamp['segment']) == segment_count
    assert [word_offset['word'] for word_offset in ts_hypotheses[0].timestamp['word']] == ts_hypotheses[0].text.split()
    assert " ".join([word_offset['word'] for word_offset in ts_hypotheses[0].timestamp['word']]) == " ".join(
        [segment_offset['segment'] for segment_offset in ts_hypotheses[0].timestamp['segment']]
    )

    assert ts_hypotheses[0].timestamp['segment'][0]['start'] == ts_hypotheses[0].timestamp['word'][0]['start']
    assert ts_hypotheses[0].timestamp['segment'][-1]['end'] == ts_hypotheses[0].timestamp['word'][-1]['end']
    assert (
        ts_hypotheses[0].timestamp['segment'][0]['start_offset']
        == ts_hypotheses[0].timestamp['word'][0]['start_offset']
    )
    assert (
        ts_hypotheses[0].timestamp['segment'][-1]['end_offset'] == ts_hypotheses[0].timestamp['word'][-1]['end_offset']
    )


@pytest.mark.unit
def test_aed_parallel_chunking(canary_1b_v2):

    audio_file = "/home/TestData/asr/longform/earnings22/sample_4469669.wav"
    # Testing on long audio file to check chunking and timestamps extraction

    hypotheses = canary_1b_v2.transcribe(audio_file, timestamps=False)
    assert len(hypotheses) == 1
    assert hypotheses[0].timestamp == []

    ts_hypotheses = canary_1b_v2.transcribe(audio_file, timestamps=True)
    assert len(ts_hypotheses) == 1

    assert ts_hypotheses[0].text == hypotheses[0].text
    assert "char" not in ts_hypotheses[0].timestamp
    assert 'word' in ts_hypotheses[0].timestamp and 'segment' in ts_hypotheses[0].timestamp
    assert len(ts_hypotheses[0].timestamp['word']) > 0
    assert len(ts_hypotheses[0].timestamp['segment']) > 0
    assert len(ts_hypotheses[0].timestamp['word']) == len(ts_hypotheses[0].text.split())

    # Monotonicity and validity of word offsets and times
    words = ts_hypotheses[0].timestamp['word']
    starts = [w['start'] for w in words]
    ends = [w['end'] for w in words]
    start_offsets = [w['start_offset'] for w in words]
    end_offsets = [w['end_offset'] for w in words]
    assert all(s <= e for s, e in zip(starts, ends))
    assert all(so <= eo for so, eo in zip(start_offsets, end_offsets))
    assert all(x <= y for x, y in zip(starts, starts[1:]))
    assert all(x <= y for x, y in zip(ends, ends[1:]))
    assert all(x <= y for x, y in zip(start_offsets, start_offsets[1:]))
    assert all(x <= y for x, y in zip(end_offsets, end_offsets[1:]))
    # Check if the transcription is correct
    assert ts_hypotheses[0].text[-25:] == 'multiple customer orders.'
    assert ts_hypotheses[0].timestamp['word'][-1] == {
        'word': 'orders.',
        'start_offset': 7477,
        'end_offset': 7481,
        'start': 598.16,
        'end': 598.48,
    }
    assert ts_hypotheses[0].text == hypotheses[0].text

    # Check that the number of words and segments are consistent
    assert [word_offset['word'] for word_offset in ts_hypotheses[0].timestamp['word']] == ts_hypotheses[0].text.split()
    assert " ".join([word_offset['word'] for word_offset in ts_hypotheses[0].timestamp['word']]) == " ".join(
        [segment_offset['segment'] for segment_offset in ts_hypotheses[0].timestamp['segment']]
    )

    assert ts_hypotheses[0].timestamp['segment'][0]['start'] == ts_hypotheses[0].timestamp['word'][0]['start']
    assert ts_hypotheses[0].timestamp['segment'][-1]['end'] == ts_hypotheses[0].timestamp['word'][-1]['end']
    assert (
        ts_hypotheses[0].timestamp['segment'][0]['start_offset']
        == ts_hypotheses[0].timestamp['word'][0]['start_offset']
    )
    assert (
        ts_hypotheses[0].timestamp['segment'][-1]['end_offset'] == ts_hypotheses[0].timestamp['word'][-1]['end_offset']
    )



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import torch

from nemo.collections.asr.models import ASRModel


class TestASRSubsamplingConvChunking:
    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_forward(self):
        asr_model = ASRModel.from_pretrained("stt_en_fastconformer_ctc_large")
        asr_model = asr_model.eval()
        asr_model.preprocessor.featurizer.dither = 0.0
        asr_model.preprocessor.featurizer.pad_to = 0

        len = 512

        input_signal_batch1 = torch.randn(size=(1, len), device=asr_model.device)
        length_batch1 = torch.randint(low=321, high=500, size=[1], device=asr_model.device)

        input_signal_batch4 = torch.randn(size=(4, len), device=asr_model.device)
        length_batch4 = torch.randint(low=321, high=500, size=[4], device=asr_model.device)

        with torch.inference_mode():
            # regular inference
            logprobs_batch1_nosplit, _, _ = asr_model.forward(
                input_signal=input_signal_batch1, input_signal_length=length_batch1
            )
            logprobs_batch4_nosplit, _, _ = asr_model.forward(
                input_signal=input_signal_batch4, input_signal_length=length_batch4
            )

            # force chunking to 2
            asr_model.change_subsampling_conv_chunking_factor(subsampling_conv_chunking_factor=2)

            # chunked inference by channels as batch is 1
            logprobs_batch1_split, _, _ = asr_model.forward(
                input_signal=input_signal_batch1, input_signal_length=length_batch1
            )
            # chunked inference by batch as it is 4 [> 1]
            logprobs_batch4_split, _, _ = asr_model.forward(
                input_signal=input_signal_batch4, input_signal_length=length_batch4
            )

        diff = torch.mean(torch.abs(logprobs_batch1_split - logprobs_batch1_nosplit))
        assert diff <= 0.2
        diff = torch.mean(torch.abs(logprobs_batch4_split - logprobs_batch4_nosplit))
        assert diff <= 0.2



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from queue import Queue
from typing import Any

import torch
from torch import Tensor


class CacheAwareContext:
    """
    Stores the cache state for the Cache-Aware models.
    """

    def __init__(
        self,
        cache_last_channel: Tensor | None = None,
        cache_last_time: Tensor | None = None,
        cache_last_channel_len: Tensor | None = None,
    ):
        """
        Args:
            cache_last_channel (Tensor | None): Last channel of the cache.
            cache_last_time (Tensor | None): Last time of the cache.
            cache_last_channel_len (Tensor | None): Last channel length of the cache.
        """
        self.cache_last_channel = cache_last_channel
        self.cache_last_time = cache_last_time
        self.cache_last_channel_len = cache_last_channel_len


class CacheAwareContextManager:
    """
    Manager class to manipulate the cached states for the Cache-Aware models.
    """

    def __init__(
        self,
        cache_aware_model: Any,
        num_slots: int,
        use_cache: bool = True,
    ):
        """
        Initialize the CacheAwareContextManager.
        Args:
            cache_aware_model (Any): Cache-Aware model object. It should have the get_initial_cache_state method.
            num_slots (int): Number of slots to use for the cache. It should be greater than or equal to the batch size.
            use_cache (bool): Whether to use the cache. Default is True. If False, the cache is disabled.
        """
        self.cache_aware_model = cache_aware_model
        # Cache aware model should have the following methods:
        if not hasattr(self.cache_aware_model, "get_initial_cache_state"):
            raise ValueError("Cache aware model should have the get_initial_cache_state method")

        self.num_slots = num_slots
        self.cache_disabled = not use_cache
        self.cache_last_channel = None
        self.cache_last_time = None
        self.cache_last_channel_len = None
        self.reset()

    def reset(self) -> None:
        """Resets the context manager"""
        if self.cache_disabled:
            return

        self.streamidx2slotidx = {}
        self.slotidx2streamidx = {}
        self.free_slots = Queue(self.num_slots)
        for i in range(self.num_slots):
            self.free_slots.put(i)
        (
            self.cache_last_channel,  # [17, B, 70, 512]
            self.cache_last_time,  # [17, B, 512, 8]
            self.cache_last_channel_len,  # B
        ) = self.cache_aware_model.get_initial_cache_state(self.num_slots)
        self.device = self.cache_last_channel.device

    def _reset_slots(self, slot_ids: list[int]) -> None:
        """
        Resets the slots for the given slot_ids
        Args:
            slot_ids: list of slot indices to reset
        """
        if self.cache_disabled:
            return

        slot_ids_tensor = torch.tensor(slot_ids, device=self.device, dtype=torch.long)
        self.cache_last_channel.index_fill_(1, slot_ids_tensor, 0.0)
        self.cache_last_time.index_fill_(1, slot_ids_tensor, 0.0)
        self.cache_last_channel_len.index_fill_(0, slot_ids_tensor, 0)

        # free the slot, so that it can be used by other streams
        # remove the stream from the mappings
        for slot_id in slot_ids:
            self.free_slots.put(slot_id)
            stream_id = self.slotidx2streamidx[slot_id]
            del self.slotidx2streamidx[slot_id]
            del self.streamidx2slotidx[stream_id]

    def update_cache(self, stream_ids: list[int], new_context: CacheAwareContext, mapping: dict) -> None:
        """
        Updates the cache for the given stream_ids with the new_context
        Args:
            stream_ids (list[int]): list of stream ids
            new_context (CacheAwareContext): new context to update corresponding to the stream_ids
            mapping (dict): mapping between the old and new slots
        """
        if self.cache_disabled:
            return

        slot_ids_list = [self.streamidx2slotidx[sid] for sid in stream_ids]
        slot_ids = torch.tensor(slot_ids_list, device=self.device, dtype=torch.long)
        tgt_slot_ids = torch.tensor(
            [mapping[sid] for sid in slot_ids_list],
            device=self.device,
            dtype=torch.long,
        )

        # In-place copy along batch/slot dimension
        self.cache_last_channel.index_copy_(1, slot_ids, new_context.cache_last_channel.index_select(1, tgt_slot_ids))
        self.cache_last_time.index_copy_(1, slot_ids, new_context.cache_last_time.index_select(1, tgt_slot_ids))
        self.cache_last_channel_len.index_copy_(
            0, slot_ids, new_context.cache_last_channel_len.index_select(0, tgt_slot_ids)
        )

    def reset_slots(self, stream_ids: list[int], eos_flags: list[bool]) -> None:
        """
        Resets the slots for the finished streams
        Args:
            stream_ids (list[int]): list of stream ids
            eos_flags (list[bool]): list of eos flags indicating whether the stream has finished
        """
        if self.cache_disabled:
            return

        if len(stream_ids) != len(eos_flags):
            raise ValueError("stream_ids and eos_flags must have the same length")

        if len(stream_ids) == 0:
            return

        # reset the slots for finished streams
        self._reset_slots([self.streamidx2slotidx[sid] for sid, eos in zip(stream_ids, eos_flags) if eos])

    def get_context(self, stream_ids: list[int]) -> tuple[CacheAwareContext, dict]:
        """
        Retrieves the context from the cache for the given stream_ids
        Args:
            stream_ids (list[int]): list of stream ids
        Returns:
            context (CacheAwareContext): context for the given stream_ids
            mapping (dict): mapping between the cache and retrieved context
        """

        if len(stream_ids) == 0 or self.cache_disabled:
            # Create a dummy context with None values
            return CacheAwareContext(), {}

        # if the stream_id is new, we need to assign a slot to it
        for stream_id in stream_ids:
            if stream_id not in self.streamidx2slotidx:
                if self.free_slots.empty():
                    raise RuntimeError("No free slots available")
                slot_idx = self.free_slots.get()
                self.streamidx2slotidx[stream_id] = slot_idx
                self.slotidx2streamidx[slot_idx] = stream_id

        # get the cache for the particular stream_ids
        slot_ids = [self.streamidx2slotidx[stream_id] for stream_id in stream_ids]
        cache_last_channel = self.cache_last_channel[:, slot_ids, :, :]
        cache_last_time = self.cache_last_time[:, slot_ids, :, :]
        cache_last_channel_len = self.cache_last_channel_len[slot_ids]

        # create a context object
        context = CacheAwareContext(
            cache_last_channel=cache_last_channel,
            cache_last_time=cache_last_time,
            cache_last_channel_len=cache_last_channel_len,
        )

        # mapping between cache and context
        mapping = dict(zip(slot_ids, range(len(slot_ids))))
        return context, mapping



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo.collections.vlm.qwen2vl.model.base import (
    Qwen2VLConfig,
    Qwen2VLModel,
    Qwen2VLVisionConfig,
    Qwen25VLVisionConfig,
)
from nemo.collections.vlm.qwen2vl.model.qwen2vl import (
    Qwen2VLConfig2B,
    Qwen2VLConfig7B,
    Qwen2VLConfig72B,
    Qwen25VLConfig3B,
    Qwen25VLConfig7B,
    Qwen25VLConfig32B,
    Qwen25VLConfig72B,
)

__all__ = [
    "Qwen2VLVisionConfig",
    "Qwen2VLConfig",
    "Qwen2VLConfig2B",
    "Qwen2VLConfig7B",
    "Qwen2VLConfig72B",
    "Qwen2VLModel",
    "Qwen25VLVisionConfig",
    "Qwen25VLConfig3B",
    "Qwen25VLConfig7B",
    "Qwen25VLConfig32B",
    "Qwen25VLConfig72B",
]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from nemo.utils import logging

COMPUTE_DTYPE_MAP = {
    'bfloat16': torch.bfloat16,
    'float16': torch.float16,
    'float32': torch.float32,
}

DEVICE_TYPES = ["cuda", "mps", "cpu"]


def setup_device(device: str, device_id: int | None, compute_dtype: str) -> tuple[str, int, torch.dtype]:
    """
    Set up the compute device for the model.

    Args:
        device (str): Requested device type ('cuda', 'mps' or 'cpu').
        device_id (int | None): Requested CUDA device ID (None for CPU or MPS).
        compute_dtype (str): Requested compute dtype.

    Returns:
        tuple(str, int, torch.dtype): Tuple of (device_string, device_id, compute_dtype) for model initialization.
    """
    device = device.strip()
    if device not in DEVICE_TYPES:
        raise ValueError(f"Invalid device type: {device}. Must be one of {DEVICE_TYPES}")

    device_id = int(device_id) if device_id is not None else 0

    # Handle CUDA devices
    if torch.cuda.is_available() and device == "cuda":
        if device_id >= torch.cuda.device_count():
            logging.warning(f"Device ID {device_id} is not available. Using GPU 0 instead.")
            device_id = 0

        compute_dtype = COMPUTE_DTYPE_MAP.get(compute_dtype, None)
        if compute_dtype is None:
            raise ValueError(
                f"Invalid compute dtype: {compute_dtype}. Must be one of {list(COMPUTE_DTYPE_MAP.keys())}"
            )

        device_str = f"cuda:{device_id}"
        return device_str, device_id, compute_dtype

    # Handle MPS devices
    if torch.backends.mps.is_available() and device == "mps":
        return "mps", -1, torch.float32

    # Handle CPU devices
    if device == "cpu":
        return "cpu", -1, torch.float32

    raise ValueError(f"Device {device} is not available.")



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from dataclasses import dataclass, field
from typing import Any, Optional, Tuple

import torch
from omegaconf import DictConfig

from nemo.utils import logging


@dataclass
class GraphIntersectDenseConfig:
    """Graph dense intersection config.
    """

    search_beam: float = 20.0
    output_beam: float = 10.0
    min_active_states: int = 30
    max_active_states: int = 10000


@dataclass
class GraphModuleConfig:
    """Config for graph modules.
    Typically used with graph losses and decoders.
    """

    topo_type: str = "default"
    topo_with_self_loops: bool = True
    token_lm: Optional[Any] = None
    intersect_pruned: bool = False
    intersect_conf: GraphIntersectDenseConfig = field(default_factory=lambda: GraphIntersectDenseConfig())
    boost_coeff: float = 0.0
    predictor_window_size: int = 0
    predictor_step_size: int = 1


class ASRK2Mixin(ABC):
    """k2 Mixin class that simplifies the construction of various models with k2-based losses.
    
    It does the following:
        -   Sets up the graph loss and decoder (methods _init_k2 and update_k2_modules).
        -   Registers external graphs, if needed.
        -   Augments forward(...) with optional graph decoding to get accurate predictions.
    """

    def _init_k2(self):
        """
        k2-related initialization implementation.

        This method is expected to run after the __init__ which sets self._cfg
        self._cfg is expected to have the attribute graph_module_cfg
        """
        if not hasattr(self, "_cfg"):
            raise ValueError("self._cfg must be set before calling _init_k2().")
        if not hasattr(self._cfg, "graph_module_cfg") or self._cfg.graph_module_cfg is None:
            raise ValueError("self._cfg.graph_module_cfg must be set and cannot be None.")
        self.graph_module_cfg = self._cfg.graph_module_cfg

        # register token_lm for MAPLoss
        criterion_type = self.graph_module_cfg.get("criterion_type", "ml")
        self.use_graph_lm = criterion_type == "map"
        if self.use_graph_lm:
            token_lm_path = self.graph_module_cfg.backend_cfg.get("token_lm", None)
            if token_lm_path is None:
                raise ValueError(
                    f"graph_module_cfg.backend_cfg.token_lm is empty. It must be set for criterion_type == `{criterion_type}`"
                )
            token_lm_path = self.register_artifact('graph_module_cfg.backend_cfg.token_lm', token_lm_path)
            self.graph_module_cfg.backend_cfg["token_lm"] = token_lm_path

        self.update_k2_modules(self.graph_module_cfg)

    def update_k2_modules(self, input_cfg: DictConfig):
        """
        Helper function to initialize or update k2 loss and transcribe_decoder.

        Args:
            input_cfg: DictConfig to take new parameters from. Schema is expected as in
                nemo.collections.asr.models.configs.k2_sequence_models_config.GraphModuleConfig
        """
        del self.loss
        if hasattr(self, "transcribe_decoder"):
            del self.transcribe_decoder

        if hasattr(self, "joint"):
            # RNNT
            num_classes = self.joint.num_classes_with_blank - 1
        else:
            # CTC, MMI, ...
            num_classes = self.decoder.num_classes_with_blank - 1
            remove_consecutive = input_cfg.backend_cfg.get("topo_with_self_loops", True) and input_cfg.backend_cfg.get(
                "topo_type", "default"
            ) not in ["forced_blank", "identity",]
            self._wer.remove_consecutive = remove_consecutive

        from nemo.collections.asr.losses.lattice_losses import LatticeLoss

        self.loss = LatticeLoss(
            num_classes=num_classes,
            reduction=self._cfg.get("ctc_reduction", "mean_batch"),
            backend="k2",
            criterion_type=input_cfg.get("criterion_type", "ml"),
            loss_type=input_cfg.get("loss_type", "ctc"),
            split_batch_size=input_cfg.get("split_batch_size", 0),
            graph_module_cfg=input_cfg.backend_cfg,
        )

        criterion_type = self.loss.criterion_type
        self.use_graph_lm = criterion_type == "map"
        transcribe_training = input_cfg.get("transcribe_training", False)
        if transcribe_training and criterion_type == "ml":
            logging.warning(
                f"""You do not need to use transcribe_training=`{transcribe_training}` 
                            with criterion_type=`{criterion_type}`. transcribe_training will be set to False."""
            )
            transcribe_training = False
        self.transcribe_training = transcribe_training
        if self.use_graph_lm:
            from nemo.collections.asr.modules.graph_decoder import ViterbiDecoderWithGraph

            self.transcribe_decoder = ViterbiDecoderWithGraph(
                num_classes=num_classes,
                backend="k2",
                dec_type="token_lm",
                return_type="1best",
                return_ilabels=True,
                output_aligned=True,
                split_batch_size=input_cfg.get("split_batch_size", 0),
                graph_module_cfg=input_cfg.backend_cfg,
            )

    def _forward_k2_post_processing(
        self, log_probs: torch.Tensor, encoded_length: torch.Tensor, greedy_predictions: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        k2-related post-processing parf of .forward()

        Args:
            log_probs: The log probabilities tensor of shape [B, T, D].
            encoded_length: The lengths of the acoustic sequence after propagation through the encoder, of shape [B].
            greedy_predictions: The greedy token predictions of the model of shape [B, T]

        Returns:
            A tuple of 3 elements -
            1) The log probabilities tensor of shape [B, T, D].
            2) The lengths of the acoustic sequence after propagation through the encoder, of shape [B].
            3) The greedy token predictions of the model of shape [B, T] (via argmax)
        """
        # greedy_predictions from .forward() are incorrect for criterion_type=`map`
        # getting correct greedy_predictions, if needed
        if self.use_graph_lm and (not self.training or self.transcribe_training):
            greedy_predictions, encoded_length, _ = self.transcribe_decoder.forward(
                log_probs=log_probs, log_probs_length=encoded_length
            )
        return log_probs, encoded_length, greedy_predictions



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List, Optional, Tuple

import torch

from nemo.core.classes.mixins import AccessMixin
from nemo.utils import logging


class InterCTCMixin:
    """Adds utilities for computing interCTC loss from https://arxiv.org/abs/2102.03216.

    To use, make sure encoder accesses ``interctc['capture_layers']``
    property in the AccessMixin and registers ``interctc/layer_output_X`` and
    ``interctc/layer_length_X`` for all layers that we want to get loss from.
    Additionally, specify the following config parameters to set up loss::

        interctc:
            # can use different values
            loss_weights: [0.3]
            apply_at_layers: [8]

    Then call

        * ``self.setup_interctc(ctc_decoder_name, ctc_loss_name, ctc_wer_name)``
          in the init method
        * ``self.add_interctc_losses`` after computing regular loss.
        * ``self.finalize_interctc_metrics(metrics, outputs, prefix="val_")``
          in the `multi_validation_epoch_end` method.
        * ``self.finalize_interctc_metrics(metrics, outputs, prefix="test_")``
          in the `multi_test_epoch_end` method.
    """

    def _process_config_values(self, loss_weights: List[float], apply_at_layers: List[int]):
        self.set_interctc_param('intermediate_loss_weights', loss_weights)
        self.set_interctc_param('apply_at_layers', apply_at_layers)
        self.set_interctc_param('main_loss_weight', 1.0 - sum(loss_weights))
        if self.get_interctc_param('main_loss_weight') <= 0.0:
            raise ValueError(
                "Make sure that sum of intermediate loss weights is < 1.0. "
                "Note that we don't do any normalization and assign "
                "remaining weight to the regular model loss. "
                "E.g., if interctc.loss_weights = [0.1, 0.3], regular "
                "loss will have weight of 0.6"
            )
        self.set_interctc_param('enabled', len(loss_weights) > 0)

        if len(apply_at_layers) != len(loss_weights):
            raise ValueError('Length of interctc.apply_at_layers has to match interctc.loss_weights')

        # setting up config for AccessMixin that will be checked in encoders to
        # log the layers we need
        AccessMixin.update_access_cfg(
            {'interctc': {'capture_layers': apply_at_layers}}, guid=getattr(self, "model_guid", None)
        )
        if hasattr(self, "propagate_model_guid"):
            self.propagate_model_guid()
        else:
            logging.warning(
                f"Not able to propagate model_guid to the submodules. Make sure to call self.propagate_model_guid() in ModelPT class."
            )

    def setup_interctc(self, decoder_name, loss_name, wer_name):
        """Sets up all interctc-specific parameters and checks config consistency.

        Caller has to specify names of attributes to perform CTC-specific WER,
        decoder and loss computation. They will be looked up in the class
        state with ``getattr``.

        The reason we get the names and look up object later is because those
        objects might change without re-calling the setup of this class. So
        we always want to look up the most up-to-date object instead of
        "caching" it here.
        """
        # registering all parameters in a dictionary to avoid conflicts with
        # main class's names
        self._interctc_params = {}
        interctc_config = self.cfg.get("interctc")
        if interctc_config is not None:
            # if interctc is in the config, we want to check that it indeed defines
            # the required keys and nothing else - that's automatically done by
            # matching with keyword arguments in self._process_config_values
            self._process_config_values(**interctc_config)
            self._interctc_params['decoder_name'] = decoder_name
            self._interctc_params['loss_name'] = loss_name
            self._interctc_params['wer_name'] = wer_name
        else:
            self.set_interctc_param('enabled', False)

    def get_interctc_param(self, param_name):
        """Either directly get parameter from ``self._interctc_params`` or
        call getattr with the corresponding name.
        """
        if param_name in ['decoder', 'loss', 'wer']:
            return getattr(self, self._interctc_params[param_name + "_name"])
        return self._interctc_params[param_name]

    def set_interctc_param(self, param_name, param_value):
        """Setting the parameter to the ``self._interctc_params`` dictionary.

        Raises an error if trying to set decoder, loss or wer as those should
        always come from the main class.
        """
        if param_name in ['decoder', 'loss', 'wer']:
            raise ValueError(
                'Cannot set "decoder", "loss" or "wer" as parameters. '
                'They are always looked up in the main class state.'
            )
        self._interctc_params[param_name] = param_value

    def _verify_setup_was_called(self):
        """Can be used to verify if setup_interctc was called."""
        if not hasattr(self, '_interctc_params'):
            raise RuntimeError(
                'self.setup_interctc(ctc_decoder_name, ctc_loss_name, ctc_wer_name) has to be '
                'called before InterCTC loss can be used!'
            )

    def is_interctc_enabled(self) -> bool:
        """Returns whether interCTC loss is enabled."""
        self._verify_setup_was_called()
        return self.get_interctc_param('enabled')

    def set_interctc_enabled(self, enabled: bool):
        """Can be used to enable/disable InterCTC manually."""
        self._verify_setup_was_called()
        if enabled:  # checking if proper config parameters were specified
            if len(self.get_interctc_param('intermediate_loss_weights')) == 0:
                raise RuntimeError(
                    'InterCTC cannot be enabled since interctc.loss_weights was not specified in the config.'
                )
            if len(self.get_interctc_param('apply_at_layers')) != len(
                self.get_interctc_param('intermediate_loss_weights')
            ):
                raise RuntimeError(
                    'InterCTC cannot be enabled, since length of "loss_weights" does not match "apply_at_layers".'
                )
        self.set_interctc_param('enabled', enabled)

    def finalize_interctc_metrics(self, metrics: Dict, outputs: List[Dict], prefix: str):
        """Finalizes InterCTC WER and loss metrics for logging purposes.

        Should be called inside ``multi_validation_epoch_end`` (with ``prefix="val_"``) or
        ``multi_test_epoch_end`` (with ``prefix="test_"``).

        Note that ``metrics`` dictionary is going to be updated in-place.
        """
        if self.is_interctc_enabled():
            for layer_idx in self.get_interctc_param('apply_at_layers'):
                # assuming that if the first batch logged the metrics, then all batches did
                if f"{prefix}inter_ctc_loss_l{layer_idx}" in outputs[0]:
                    loss = torch.stack([x[f"{prefix}inter_ctc_loss_l{layer_idx}"] for x in outputs]).mean()
                    metrics["log"][f"{prefix}inter_ctc_loss_l{layer_idx}"] = loss

                if f"{prefix}inter_wer_num_l{layer_idx}" in outputs[0]:
                    wer_num = torch.stack([x[f"{prefix}inter_wer_num_l{layer_idx}"] for x in outputs]).sum()
                    wer_denom = torch.stack([x[f"{prefix}inter_wer_denom_l{layer_idx}"] for x in outputs]).sum()
                    metrics["log"][f"{prefix}inter_wer_l{layer_idx}"] = wer_num / wer_denom

            if f"{prefix}final_loss" in outputs[0]:
                metrics["log"][f"{prefix}final_loss"] = torch.stack([x[f"{prefix}final_loss"] for x in outputs]).mean()

    def get_captured_interctc_tensors(self) -> List[Tuple[torch.Tensor, torch.Tensor]]:
        """Returns a list of captured tensors from encoder: tuples of (output, length).

        Will additionally apply ``ctc_decoder`` to the outputs.
        """
        if not self.is_interctc_enabled():
            return []

        # note that we have a loop here, because tensors can be defined from
        # submodules of encoder (e.g., that's the case in Jasper)
        total_registry = {}
        for module_registry in AccessMixin.get_module_registry(self.encoder).values():
            for key in module_registry:
                if key.startswith("interctc/") and key in total_registry:
                    raise RuntimeError(f"layer {key} has been logged multiple times!")
            total_registry.update(module_registry)
        # if intermediate_loss_weights was set, the encoder has to register
        # interctc/layer_output_X and interctc/layer_length_X tensors.
        # We need to apply decoder to each of them and compute CTC loss.
        captured_tensors = []
        for layer_idx in self.get_interctc_param('apply_at_layers'):
            try:
                layer_outputs = total_registry[f"interctc/layer_output_{layer_idx}"]
                layer_lengths = total_registry[f"interctc/layer_length_{layer_idx}"]
            except KeyError:
                raise RuntimeError(
                    f"Intermediate layer {layer_idx} was not captured! "
                    "Check if length of model.encoder.captured_layer_outputs matches "
                    "length of model.intermediate_loss_weights properties."
                )
            if len(layer_outputs) > 1 or len(layer_lengths) > 1:
                raise RuntimeError(
                    "Make sure encoder.forward is called exactly one time before interCTC loss is computed."
                )
            captured_tensors.append(
                (self.get_interctc_param('decoder')(encoder_output=layer_outputs[0]), layer_lengths[0])
            )
        return captured_tensors

    def add_interctc_losses(
        self,
        loss_value: torch.Tensor,
        transcript: torch.Tensor,
        transcript_len: torch.Tensor,
        compute_wer: bool,
        compute_loss: bool = True,
        log_wer_num_denom: bool = False,
        log_prefix: str = "",
    ) -> Tuple[Optional[torch.Tensor], Dict]:
        """Adding interCTC losses if required.

        Will also register loss/wer metrics in the returned dictionary.

        Args:
            loss_value (torch.Tensor): regular loss tensor (will add interCTC loss to it).
            transcript (torch.Tensor): current utterance transcript.
            transcript_len (torch.Tensor): current utterance transcript length.
            compute_wer (bool): whether to compute WER for the current utterance.
                Should typically be True for validation/test and only True for
                training if current batch WER should be logged.
            compute_loss (bool): whether to compute loss for the current utterance.
                Should always be True in training and almost always True in
                validation, unless all other losses are disabled as well.
                Defaults to True.
            log_wer_num_denom (bool): if True, will additionally log WER num/denom
                in the returned metrics dictionary. Should always be True for
                validation/test to allow correct metrics aggregation. Should
                always be False for training. Defaults to False.
            log_prefix (str): prefix added to all log values. Should be ``""`` for
                training and ``"val_"`` for validation. Defaults to "".

        Returns:
            tuple[Optional[torch.Tensor], Dict]: tuple of new loss tensor and dictionary with logged metrics.
        """
        if not self.is_interctc_enabled() or not AccessMixin.is_access_enabled(getattr(self, "model_guid", None)):
            return loss_value, {}
        metrics = {}
        if compute_loss:
            metrics[f"{log_prefix}final_loss"] = loss_value
        else:
            loss_value = None
        captured_tensors = self.get_captured_interctc_tensors()

        if compute_loss:
            loss_value *= self.get_interctc_param('main_loss_weight')

        for layer_idx, intermediate_result, loss_weight in zip(
            self.get_interctc_param('apply_at_layers'),
            captured_tensors,
            self.get_interctc_param('intermediate_loss_weights'),
        ):
            if compute_loss:
                inter_loss_value = self.get_interctc_param('loss')(
                    log_probs=intermediate_result[0],
                    targets=transcript,
                    target_lengths=transcript_len,
                    input_lengths=intermediate_result[1],
                )
                metrics[f"{log_prefix}inter_ctc_loss_l{layer_idx}"] = inter_loss_value.detach()
                loss_value += inter_loss_value * loss_weight
            if compute_wer:
                self.get_interctc_param('wer').update(
                    predictions=intermediate_result[0],
                    targets=transcript,
                    targets_lengths=transcript_len,
                    predictions_lengths=intermediate_result[1],
                )
                wer, wer_num, wer_denom = self.get_interctc_param('wer').compute()
                self.get_interctc_param('wer').reset()
                metrics.update({f'{log_prefix}inter_wer_l{layer_idx}': wer})
                if log_wer_num_denom:
                    metrics.update(
                        {
                            f'{log_prefix}inter_wer_num_l{layer_idx}': wer_num,
                            f'{log_prefix}inter_wer_denom_l{layer_idx}': wer_denom,
                        }
                    )

        # return total loss and dictionary of metrics
        return loss_value, metrics



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, List, Optional

import numpy as np
import torch
import torch.nn.functional as F
from omegaconf import DictConfig, ListConfig

from nemo.collections.asr.parts.context_biasing.biasing_multi_model import GPUBiasingMultiModel
from nemo.collections.asr.parts.submodules.ngram_lm import NGramGPULanguageModel
from nemo.collections.asr.parts.submodules.transducer_decoding.label_looping_base import (
    BatchedLabelLoopingState,
    GreedyBatchedLabelLoopingComputerBase,
    LabelLoopingStateItem,
    SeparateGraphsLabelLooping,
)
from nemo.collections.asr.parts.utils import rnnt_utils
from nemo.collections.asr.parts.utils.asr_confidence_utils import ConfidenceMethodMixin
from nemo.core.utils.cuda_python_utils import NeMoCUDAPythonException, cu_call, run_nvrtc, with_conditional_node
from nemo.utils import logging

try:
    from cuda.bindings import runtime as cudart

    HAVE_CUDA_PYTHON = True
except ImportError:
    HAVE_CUDA_PYTHON = False


class LabelLoopingState:
    """
    State for Loop Labels algorithm. Used only with CUDA graphs.
    In initialization phase it is possible to assign values (tensors) to the state.
    For algorithm code the storage should be reused (prefer copy data instead of assigning tensors).
    """

    max_time: int  # maximum length of internal storage for time dimension
    batch_size: int  # (maximum) length of internal storage for batch dimension
    device: torch.device  # device to store preallocated tensors

    model_durations: torch.Tensor

    encoder_output_projected: torch.Tensor  # projected output from the encoder for decoding algorithm
    encoder_output_length: torch.Tensor  # length of the (projected) output from the encoder

    labels: torch.Tensor  # storage for current labels
    scores: torch.Tensor  # storage for current scores
    durations: torch.Tensor  # storage for current predicted durations

    batch_indices: torch.Tensor  # indices of elements in batch (constant, range [0, batch_size-1])

    time_indices: torch.Tensor  # current time indices for each element in batch
    safe_time_indices: torch.Tensor  # current time indices, but guaranteed to be < encoder_output_length
    time_indices_current_labels: torch.Tensor  # time indices for found labels (corresponding to `labels` field)
    last_timesteps: torch.Tensor  # indices of the last timesteps for each element (encoder_output_length - 1)

    active_mask: torch.Tensor  # mask for active hypotheses (the decoding is finished for the utterance if it is False)
    advance_mask: torch.Tensor  # mask for "advancing" hypotheses (blank is found for the element on the current step)
    blank_mask: torch.Tensor  # if the element is blank
    active_mask_prev: torch.Tensor  # if the element was active on the previous step
    found_labels_mask: torch.Tensor  # mask for found labels (non-blank)

    active_mask_any: torch.Tensor  # 0-dim bool tensor, condition for outer loop ('any element is still active')
    advance_mask_any: torch.Tensor  # 0-dim bool tensor, condition for inner loop ('should advance any index')

    decoder_state: Any  # current decoder state
    decoder_output: torch.Tensor  # output from the decoder (projected)

    decoder_state_after_sos: Any  # decoder state after _SOS symbol (for initialization)
    decoder_output_after_sos: (
        torch.Tensor
    )  # output from the decoder (projected) after _SOS symbol (for initialization)

    batched_hyps: rnnt_utils.BatchedHyps  # batched hypotheses - decoding result
    alignments: Optional[rnnt_utils.BatchedAlignments] = None  # batched alignments

    # for fusion models
    fusion_states_list: list[torch.Tensor]
    fusion_states_candidates_list: list[torch.Tensor]
    fusion_scores_list: list[torch.Tensor]
    multi_biasing_ids: torch.Tensor

    def __init__(
        self,
        batch_size: int,
        max_time: int,
        encoder_dim: int,
        max_symbols: int,
        device: torch.device,
        float_dtype: torch.dtype,
        logits_dim: int,
        preserve_alignments=False,
        preserve_frame_confidence=False,
        include_duration_confidence: bool = False,
        include_duration: bool = False,
    ):
        """

        Args:
            batch_size: batch size for encoder output storage
            max_time: maximum time for encoder output storage
            encoder_dim: last dimension for encoder output storage (projected encoder output)
            max_symbols: max symbols per step (to avoid infinite looping and pre-allocate storage)
            device: device to store tensors
            float_dtype: default float dtype for tensors (should match projected encoder output)
            logits_dim: output dimension for Joint
            preserve_alignments: if alignments are needed
            preserve_frame_confidence: if frame confidence is needed
            include_duration_confidence: if duration confidence is needed to be added to the frame confidence
            include_duration: if predicted token durations are needed to be added to the Hypothesis object
        """
        self.device = device
        self.float_dtype = float_dtype
        self.batch_size = batch_size
        self.max_time = max_time
        self.include_duration = include_duration

        self.encoder_output_projected = torch.zeros(
            (self.batch_size, self.max_time, encoder_dim),
            dtype=float_dtype,
            device=self.device,
        )
        self.encoder_output_length = torch.zeros((self.batch_size,), dtype=torch.long, device=self.device)

        self.labels = torch.zeros([self.batch_size], dtype=torch.long, device=self.device)
        self.scores = torch.zeros([self.batch_size], dtype=float_dtype, device=self.device)

        # indices of elements in batch (constant)
        self.batch_indices = torch.arange(self.batch_size, dtype=torch.long, device=self.device)

        self.time_indices = torch.zeros_like(self.batch_indices)
        self.safe_time_indices = torch.zeros_like(self.batch_indices)
        self.time_indices_current_labels = torch.zeros_like(self.time_indices)
        self.last_timesteps = torch.zeros_like(self.time_indices)
        self.durations = torch.zeros([self.batch_size], dtype=torch.long, device=self.device)

        self.active_mask = torch.zeros([self.batch_size], dtype=torch.bool, device=self.device)
        self.advance_mask = torch.zeros_like(self.active_mask)
        self.blank_mask = torch.zeros_like(self.active_mask)
        self.active_mask_prev = torch.zeros_like(self.active_mask)
        self.found_labels_mask = torch.zeros_like(self.active_mask)

        self.active_mask_any = torch.tensor(True, device=self.device, dtype=torch.bool)
        self.advance_mask_any = torch.tensor(True, device=self.device, dtype=torch.bool)

        self.multi_biasing_ids = torch.full([self.batch_size], fill_value=-1, dtype=torch.long, device=self.device)
        self.fusion_states_list = []
        self.fusion_states_candidates_list = []
        self.fusion_scores_list = []
        self.batched_hyps = rnnt_utils.BatchedHyps(
            batch_size=self.batch_size,
            init_length=self.max_time * max_symbols,
            device=self.device,
            float_dtype=float_dtype,
            is_with_durations=include_duration,
        )
        if preserve_alignments or preserve_frame_confidence:
            self.alignments = rnnt_utils.BatchedAlignments(
                batch_size=batch_size,
                logits_dim=logits_dim,
                init_length=max_time * (max_symbols + 1),
                device=self.device,
                float_dtype=self.float_dtype,
                store_alignments=preserve_alignments,
                store_frame_confidence=preserve_frame_confidence,
                with_duration_confidence=include_duration_confidence,
            )
        else:
            self.alignments = None

    def need_reinit(self, encoder_output_projected: torch.Tensor) -> bool:
        """Check if need to reinit state: larger batch_size/max_time, or new device"""
        return (
            self.batch_size < encoder_output_projected.shape[0]
            or self.max_time < encoder_output_projected.shape[1]
            or self.device.index != encoder_output_projected.device.index
        )


class GreedyBatchedTDTLabelLoopingComputer(GreedyBatchedLabelLoopingComputerBase, ConfidenceMethodMixin):
    """
    Label-Looping algorithm implementation https://arxiv.org/abs/2406.06220 for optimized batched greedy decoding.
    Iterates over labels, on each step finding the next non-blank label
    (evaluating Joint multiple times in inner loop); It uses a minimal possible amount of calls
    to prediction network (with maximum possible batch size),
    which makes it especially useful for scaling the prediction network.
    During decoding all active hypotheses ("texts") have the same lengths.
    """

    INITIAL_MAX_TIME = 375  # initial max time, used to init state for Cuda graphs
    CUDA_PROGRAM_NAME = b"while_label_looping_conditional_tdt.cu"

    separate_graphs: Optional[SeparateGraphsLabelLooping]
    full_graph: Optional[torch.cuda.CUDAGraph]
    state: Optional[LabelLoopingState]
    fusion_models: Optional[List[NGramGPULanguageModel]]

    def __init__(
        self,
        decoder,
        joint,
        blank_index: int,
        durations: list[int] | ListConfig[int],
        max_symbols_per_step: Optional[int] = None,
        preserve_alignments=False,
        preserve_frame_confidence=False,
        include_duration: bool = False,
        include_duration_confidence: bool = False,
        confidence_method_cfg: Optional[DictConfig] = None,
        allow_cuda_graphs: bool = True,
        fusion_models: Optional[List[NGramGPULanguageModel]] = None,
        fusion_models_alpha: Optional[List[float]] = None,
        enable_per_stream_biasing: bool = False,
    ):
        """
        Init method.
        Args:
            decoder: Prediction network from RNN-T
            joint: Joint module from RNN-T
            blank_index: index of blank symbol
            durations: list of TDT durations, e.g., [0, 1, 2, 4, 8]
            max_symbols_per_step: max symbols to emit on each step (to avoid infinite looping)
            preserve_alignments: if alignments are needed
            preserve_frame_confidence: if frame confidence is needed
            include_duration: if predicted token durations are needed to be added to the Hypothesis object
            include_duration_confidence: if duration confidence is needed to be added to the frame confidence
            confidence_method_cfg: config for the confidence
            fusion_models: list of fusion models (ngram_lm_model and boosting_tree_model)
            fusion_models_alpha: list of fusion model weights (ngram_lm_alpha and boosting_tree_alpha)
            enable_per_stream_biasing: enable multi-biasing model for per-stream customization
        """
        super().__init__()
        self.decoder = decoder
        self.joint = joint
        # keep durations on CPU to avoid side effects in multi-gpu environment
        self.durations = torch.tensor(list(durations), device="cpu").to(torch.long)
        self._blank_index = blank_index
        self.max_symbols = max_symbols_per_step
        self.preserve_alignments = preserve_alignments
        self.preserve_frame_confidence = preserve_frame_confidence
        self.preserve_alignments = preserve_alignments or preserve_frame_confidence
        self.include_duration = include_duration
        self.include_duration_confidence = include_duration_confidence
        self._SOS = self._blank_index
        self._init_confidence_method(confidence_method_cfg=confidence_method_cfg)
        assert self._SOS == self._blank_index  # "blank as pad" algorithm only

        self.fusion_models = fusion_models or []
        self.fusion_models_alpha = fusion_models_alpha or []

        self.biasing_multi_model = (
            GPUBiasingMultiModel(vocab_size=self._blank_index, reallocation_callback_fn=self.reset_cuda_graphs_state)
            if enable_per_stream_biasing
            else None
        )

        if allow_cuda_graphs:
            for fusion_model in self._all_fusion_models():
                if not fusion_model.compatible_with_cuda_graphs():
                    logging.warning(
                        "Fusion model used that is incompatible with CUDA graphs."
                        " Switching off CUDA graphs, decoding may be slow."
                    )
                    allow_cuda_graphs = False
                    break

        self.allow_cuda_graphs = allow_cuda_graphs

        self.state = None
        self.full_graph = None
        self.separate_graphs = None

        self.cuda_graphs_mode = None
        self.cuda_graphs_allow_fallback = True
        self.maybe_enable_cuda_graphs()

    def reset_cuda_graphs_state(self):
        """Reset state to release memory (for CUDA graphs implementations)"""
        self.state = None
        self.full_graph = None
        self.separate_graphs = None

    def _get_frame_confidence(self, logits: torch.Tensor, num_durations: int) -> Optional[torch.Tensor]:
        float_dtype = logits.dtype
        return (
            torch.stack(
                (
                    self._get_confidence_tensor(F.log_softmax(logits[:, :-num_durations], dim=-1)).to(
                        dtype=float_dtype
                    ),
                    self._get_confidence_tensor(F.log_softmax(logits[:, -num_durations:], dim=-1)).to(
                        dtype=float_dtype
                    ),
                ),
                dim=-1,
            )
            if self.include_duration_confidence
            else (
                self._get_confidence_tensor(F.log_softmax(logits[:, :-num_durations], dim=-1)).to(dtype=float_dtype)
                if self.preserve_frame_confidence
                else None
            )
        )

    def torch_impl(
        self,
        encoder_output: torch.Tensor,
        encoder_output_length: torch.Tensor,
        prev_batched_state: Optional[BatchedLabelLoopingState] = None,
        multi_biasing_ids: Optional[torch.Tensor] = None,
    ) -> tuple[rnnt_utils.BatchedHyps, Optional[rnnt_utils.BatchedAlignments], BatchedLabelLoopingState]:
        """
        Pure PyTorch implementation

        Args:
            encoder_output: output from the encoder
            encoder_output_length: lengths of the utterances in `encoder_output`
            prev_batched_state: previous batched decoding state
            multi_biasing_ids: optional tensor [Batch] with ids of multi-biasing models
        """
        batch_size, max_time, _unused = encoder_output.shape
        device = encoder_output.device
        self._move_fusion_models_to_device(device=device)
        if self.biasing_multi_model is not None and multi_biasing_ids is None:
            multi_biasing_ids = torch.full([batch_size], fill_value=-1, dtype=torch.long, device=device)

        # do not recalculate joint projection, project only once
        encoder_output_projected = self.joint.project_encoder(encoder_output)
        float_dtype = encoder_output_projected.dtype

        # init output structures: BatchedHyps (for results), BatchedAlignments + last decoder state
        # init empty batched hypotheses
        batched_hyps = rnnt_utils.BatchedHyps(
            batch_size=batch_size,
            init_length=max_time * self.max_symbols if self.max_symbols is not None else max_time,
            device=device,
            float_dtype=float_dtype,
            is_with_durations=self.include_duration,
        )
        # init alignments if necessary
        use_alignments = self.preserve_alignments or self.preserve_frame_confidence
        # always use alignments variable - for torch.jit adaptation, but keep it as minimal as possible
        alignments = rnnt_utils.BatchedAlignments(
            batch_size=batch_size,
            logits_dim=self.joint.num_classes_with_blank,
            init_length=max_time * 2 if use_alignments else 1,  # blank for each timestep + text tokens
            device=device,
            float_dtype=float_dtype,
            store_alignments=self.preserve_alignments,
            store_frame_confidence=self.preserve_frame_confidence,
            with_duration_confidence=self.include_duration_confidence,
        )

        # durations
        model_durations = self.durations.to(device, non_blocking=True)
        num_durations = model_durations.shape[0]

        # indices of elements in batch (constant)
        batch_indices = torch.arange(batch_size, dtype=torch.long, device=device)

        # time indices
        last_timesteps = torch.maximum(encoder_output_length - 1, torch.zeros_like(encoder_output_length))
        time_indices = (
            torch.zeros_like(batch_indices) if prev_batched_state is None else prev_batched_state.time_jumps.clone()
        )
        safe_time_indices = torch.minimum(time_indices, last_timesteps)  # time indices, guaranteed to be < out_len
        time_indices_current_labels = torch.zeros_like(time_indices)

        # masks for utterances in batch
        active_mask: torch.Tensor = time_indices < encoder_output_length
        active_mask_prev = active_mask.clone()
        advance_mask = torch.empty_like(active_mask)

        if prev_batched_state is None:
            # initial state, needed for torch.jit to compile (cannot handle None)
            state = self.decoder.initialize_state(encoder_output_projected)
            # last found labels - initially <SOS> (<blank>) symbol
            labels = torch.full_like(batch_indices, fill_value=self._SOS)
            decoder_output, state, *_ = self.decoder.predict(
                labels.unsqueeze(1), state, add_sos=False, batch_size=batch_size
            )
            decoder_output = self.joint.project_prednet(decoder_output)  # do not recalculate joint projection

            # fusion models
            fusion_states_list = []
            for fusion_model in self._all_fusion_models():
                fusion_states_list.append(fusion_model.get_init_states(batch_size=batch_size, bos=True))
        else:
            decoder_output = prev_batched_state.predictor_outputs
            state = prev_batched_state.predictor_states
            fusion_states_list = prev_batched_state.fusion_states_list

        fusion_scores_list, fusion_states_candidates_list = [], []

        # loop while there are active utterances
        while active_mask.any():
            # stage 1: get joint output, iteratively seeking for non-blank labels
            # blank label in `labels` tensor means "end of hypothesis" (for this index)
            active_mask_prev.copy_(active_mask)

            # stage 1.1: get first joint output
            logits = (
                self.joint.joint_after_projection(
                    encoder_output_projected[batch_indices, safe_time_indices].unsqueeze(1),
                    decoder_output,
                )
                .squeeze(1)
                .squeeze(1)
            )
            scores, labels = logits[:, :-num_durations].max(dim=-1)

            if self.has_fusion_models():
                fusion_scores_list, fusion_states_candidates_list = self.advance_fusion_models(
                    fusion_states_list=fusion_states_list,
                    multi_biasing_ids=multi_biasing_ids,
                    float_dtype=float_dtype,
                )
                logits_with_fusion = logits.clone()
                for fusion_scores in fusion_scores_list:
                    logits_with_fusion[:, : -num_durations - 1] += fusion_scores

                # get max scores and labels without blank
                fusion_scores_max, fusion_labels_max = logits_with_fusion[:, : -num_durations - 1].max(dim=-1)
                # preserve "blank" / "non-blank" category
                torch.where(labels == self._blank_index, labels, fusion_labels_max, out=labels)
                torch.where(labels == self._blank_index, scores, fusion_scores_max, out=scores)

            jump_durations_indices = logits[:, -num_durations:].argmax(dim=-1)
            durations = model_durations[jump_durations_indices]

            # search for non-blank labels using joint, advancing time indices for blank labels
            # checking max_symbols is not needed, since we already forced advancing time indices for such cases
            blank_mask = labels == self._blank_index
            # for blank labels force duration >= 1
            durations.masked_fill_(torch.logical_and(durations == 0, blank_mask), 1)
            time_indices_current_labels.copy_(time_indices)
            if use_alignments:
                alignments.add_results_masked_(
                    active_mask=active_mask,
                    time_indices=time_indices_current_labels,
                    logits=logits if self.preserve_alignments else None,
                    labels=labels if self.preserve_alignments else None,
                    confidence=self._get_frame_confidence(logits=logits, num_durations=num_durations),
                )

            # advance_mask is a mask for current batch for searching non-blank labels;
            # each element is True if non-blank symbol is not yet found AND we can increase the time index
            time_indices += durations * active_mask
            torch.minimum(time_indices, last_timesteps, out=safe_time_indices)
            torch.less(time_indices, encoder_output_length, out=active_mask)
            torch.logical_and(active_mask, blank_mask, out=advance_mask)

            # stage 1.2: inner loop - find next non-blank labels (if exist)
            while advance_mask.any():
                # same as: time_indices_current_labels[advance_mask] = time_indices[advance_mask], but non-blocking
                # store current time indices to use further for storing the results
                torch.where(advance_mask, time_indices, time_indices_current_labels, out=time_indices_current_labels)
                logits = (
                    self.joint.joint_after_projection(
                        encoder_output_projected[batch_indices, safe_time_indices].unsqueeze(1),
                        decoder_output,
                    )
                    .squeeze(1)
                    .squeeze(1)
                )
                # get labels (greedy) and scores from current logits, replace labels/scores with new
                # labels[advance_mask] are blank, and we are looking for non-blank labels
                more_scores, more_labels = logits[:, :-num_durations].max(dim=-1)

                if self.has_fusion_models():
                    logits_with_fusion = logits.clone()
                    for fusion_scores in fusion_scores_list:
                        logits_with_fusion[:, : -num_durations - 1] += fusion_scores
                    # get max scores and labels without blank
                    more_scores_w_fusion, more_labels_w_fusion = logits_with_fusion[:, : -num_durations - 1].max(
                        dim=-1
                    )
                    # preserve "blank" / "non-blank" category
                    torch.where(more_labels == self._blank_index, more_labels, more_labels_w_fusion, out=more_labels)

                # same as: labels[advance_mask] = more_labels[advance_mask], but non-blocking
                torch.where(advance_mask, more_labels, labels, out=labels)
                # same as: scores[advance_mask] = more_scores[advance_mask], but non-blocking
                torch.where(advance_mask, more_scores, scores, out=scores)
                jump_durations_indices = logits[:, -num_durations:].argmax(dim=-1)
                durations = model_durations[jump_durations_indices]

                if use_alignments:
                    alignments.add_results_masked_(
                        active_mask=advance_mask,
                        time_indices=time_indices_current_labels,
                        logits=logits if self.preserve_alignments else None,
                        labels=more_labels if self.preserve_alignments else None,
                        confidence=self._get_frame_confidence(logits=logits, num_durations=num_durations),
                    )

                blank_mask = labels == self._blank_index
                # for blank labels force duration >= 1
                durations.masked_fill_(torch.logical_and(durations == 0, blank_mask), 1)
                # same as time_indices[advance_mask] += durations[advance_mask], but non-blocking
                torch.where(advance_mask, time_indices + durations, time_indices, out=time_indices)
                torch.minimum(time_indices, last_timesteps, out=safe_time_indices)
                torch.less(time_indices, encoder_output_length, out=active_mask)
                torch.logical_and(active_mask, blank_mask, out=advance_mask)

            # NB: difference between RNN-T and TDT here, at the end of utterance:
            # For RNN-T, if we found a non-blank label, the utterance is active (need to find blank to stop decoding)
            # For TDT, we could find a non-blank label, add duration, and the utterance may become inactive
            found_labels_mask = torch.logical_and(active_mask_prev, labels != self._blank_index)
            # store hypotheses
            if self.max_symbols is not None:
                # pre-allocated memory, no need for checks
                batched_hyps.add_results_masked_no_checks_(
                    active_mask=found_labels_mask,
                    labels=labels,
                    time_indices=time_indices_current_labels,
                    scores=scores,
                    token_durations=durations if self.include_duration else None,
                )
            else:
                # auto-adjusted storage
                batched_hyps.add_results_masked_(
                    active_mask=found_labels_mask,
                    labels=labels,
                    time_indices=time_indices_current_labels,
                    scores=scores,
                    token_durations=durations if self.include_duration else None,
                )

            # stage 3: get decoder (prediction network) output with found labels
            # NB: if active_mask is False, this step is redundant;
            # but such check will require device-to-host synchronization, so we avoid it
            # preserve state/decoder_output for inactive elements
            prev_state = state
            prev_decoder_output = decoder_output
            decoder_output, state, *_ = self.decoder.predict(
                labels.unsqueeze(1), state, add_sos=False, batch_size=batch_size
            )
            decoder_output = self.joint.project_prednet(decoder_output)  # do not recalculate joint projection

            # preserve correct states/outputs for inactive elements
            self.decoder.batch_replace_states_mask(
                src_states=prev_state,
                dst_states=state,
                mask=~found_labels_mask,
            )
            torch.where(
                found_labels_mask.unsqueeze(-1).unsqueeze(-1), decoder_output, prev_decoder_output, out=decoder_output
            )

            for fusion_states, fusion_states_candidates in zip(fusion_states_list, fusion_states_candidates_list):
                torch.where(
                    found_labels_mask,
                    fusion_states_candidates[batch_indices, labels * found_labels_mask],
                    fusion_states,
                    out=fusion_states,
                )

            # stage 4: to avoid infinite looping, go to the next frame after max_symbols emission
            if self.max_symbols is not None:
                # if labels are non-blank (not end-of-utterance), check that last observed timestep with label:
                # if it is equal to the current time index, and number of observations is >= max_symbols, force blank
                force_blank_mask = torch.logical_and(
                    active_mask,
                    torch.logical_and(
                        torch.logical_and(
                            labels != self._blank_index,
                            batched_hyps.last_timestamp_lasts >= self.max_symbols,
                        ),
                        batched_hyps.last_timestamp == time_indices,
                    ),
                )
                time_indices += force_blank_mask  # emit blank => advance time indices
                # update safe_time_indices, non-blocking
                torch.minimum(time_indices, last_timesteps, out=safe_time_indices)
                # same as: active_mask = time_indices < encoder_output_length
                torch.less(time_indices, encoder_output_length, out=active_mask)

        # fix timestamps for iterative decoding
        if prev_batched_state is not None:
            batched_hyps.timestamps += prev_batched_state.decoded_lengths.unsqueeze(1)
            if use_alignments:
                alignments.timestamps += prev_batched_state.decoded_lengths.unsqueeze(1)
        # NB: last labels can not exist (nothing decoded on this step).
        # return the last labels from the previous state in this case
        last_labels = batched_hyps.get_last_labels(pad_id=self._SOS)
        decoding_state = BatchedLabelLoopingState(
            predictor_states=state,
            predictor_outputs=decoder_output,
            labels=(
                torch.where(last_labels == self._SOS, prev_batched_state.labels, last_labels)
                if prev_batched_state is not None
                else last_labels
            ),
            decoded_lengths=(
                encoder_output_length.clone()
                if prev_batched_state is None
                else encoder_output_length + prev_batched_state.decoded_lengths
            ),
            fusion_states_list=fusion_states_list,
            time_jumps=time_indices - encoder_output_length,
        )
        if use_alignments:
            return batched_hyps, alignments, decoding_state
        return batched_hyps, None, decoding_state

    def _get_decoding_state_item_after_sos(self, device: torch.device | str) -> LabelLoopingStateItem:
        """Get decoding state item after <SOS> symbol, used for initialization from empty hypotheses."""
        batched_state = self._get_batched_decoding_state_after_sos(device=device, batch_size=1)
        return self.split_batched_state(batched_state)[0]

    def _get_batched_decoding_state_after_sos(
        self, device: torch.device | str, batch_size: int
    ) -> BatchedLabelLoopingState:
        """Get batched decoding state after <SOS> symbol, used for initialization from empty hypotheses."""
        labels = torch.full([batch_size], fill_value=self._SOS, dtype=torch.long, device=device)
        decoder_output, state, *_ = self.decoder.predict(
            labels.unsqueeze(1), None, add_sos=False, batch_size=batch_size
        )
        decoder_output = self.joint.project_prednet(decoder_output)  # do not recalculate joint projection
        state = BatchedLabelLoopingState(
            predictor_states=state,
            predictor_outputs=decoder_output,
            labels=labels,
            decoded_lengths=torch.zeros([batch_size], dtype=torch.long, device=device),
            fusion_states_list=(
                [
                    fusion_model.get_init_states(batch_size=batch_size, bos=True).to(device)
                    for fusion_model in self._all_fusion_models()
                ]
            ),
            time_jumps=torch.zeros([batch_size], dtype=torch.long, device=device),
        )
        return state

    def reset_state_by_mask(self, state: BatchedLabelLoopingState, mask: torch.Tensor) -> BatchedLabelLoopingState:
        """
        Reset state for masked elements in the batched state.
        This is used to reset state for elements that are not active anymore to start a new decoding session.

        Args:
            state: batched decoding state
            mask: mask for elements to reset
        """
        state_after_sos = self._get_batched_decoding_state_after_sos(
            device=state.predictor_outputs.device, batch_size=state.labels.shape[0]
        )
        self.decoder.batch_replace_states_mask(
            src_states=state_after_sos.predictor_states, dst_states=state.predictor_states, mask=mask
        )
        torch.where(
            mask[:, None, None],
            state_after_sos.predictor_outputs,
            state.predictor_outputs,
            out=state.predictor_outputs,
        )
        torch.where(mask, state_after_sos.labels, state.labels, out=state.labels)
        torch.where(mask, state_after_sos.decoded_lengths, state.decoded_lengths, out=state.decoded_lengths)
        for fusion_idx, fusion_states in enumerate(state.fusion_states_list):
            torch.where(
                mask,
                state_after_sos.fusion_states_list[fusion_idx],
                fusion_states,
                out=state.fusion_states_list[fusion_idx],
            )
        torch.where(mask, state_after_sos.time_jumps, state.time_jumps, out=state.time_jumps)
        return state

    def split_batched_state(self, state: BatchedLabelLoopingState) -> list[LabelLoopingStateItem]:
        """
        Split batched state into list of items, each item contains state for one hypothesis.
        This is used to pass state between invocations of the algorithm.

        Args:
            state: batched decoding state
        """
        state_items: list[LabelLoopingStateItem] = []
        for i, predictor_state in enumerate(self.decoder.batch_split_states(state.predictor_states)):
            state_items.append(
                LabelLoopingStateItem(
                    predictor_state=predictor_state,
                    predictor_output=state.predictor_outputs[i],
                    label=state.labels[i],
                    decoded_length=state.decoded_lengths[i],
                    fusion_state_list=([fusion_state[i] for fusion_state in state.fusion_states_list]),
                    time_jump=state.time_jumps[i],
                )
            )
        return state_items

    def merge_to_batched_state(self, state_items: list[LabelLoopingStateItem | None]) -> BatchedLabelLoopingState:
        """
        Merge list of items into batched state, each item contains state for one hypothesis.
        This is used to pass state between invocations of the algorithm.

        Args:
            state_items: list of items to merge
        """
        if any(item is None for item in state_items):
            not_none_item = next(item for item in state_items if item is not None)
            assert not_none_item is not None
            device = not_none_item.predictor_output.device
            start_item = self._get_decoding_state_item_after_sos(device=device)
            for i, item in enumerate(state_items):
                if item is None:
                    state_items[i] = start_item

        fusion_states_list = []
        for fusion_idx in range(len(self._all_fusion_models())):
            fusion_states_list.append(torch.stack([item.fusion_state_list[fusion_idx] for item in state_items]))

        batched_state = BatchedLabelLoopingState(
            predictor_states=self.decoder.batch_unsplit_states([item.predictor_state for item in state_items]),
            predictor_outputs=torch.stack([item.predictor_output for item in state_items]),
            labels=torch.stack([item.label for item in state_items]),
            decoded_lengths=torch.stack([item.decoded_length for item in state_items]),
            fusion_states_list=fusion_states_list,
            time_jumps=torch.stack([item.time_jump for item in state_items]),
        )
        return batched_state

    def cuda_graphs_impl(
        self,
        encoder_output: torch.Tensor,
        encoder_output_length: torch.Tensor,
        prev_batched_state: Optional[BatchedLabelLoopingState] = None,
        multi_biasing_ids: Optional[torch.Tensor] = None,
    ) -> tuple[rnnt_utils.BatchedHyps, Optional[rnnt_utils.BatchedAlignments], BatchedLabelLoopingState]:
        """
        Implementation with CUDA graphs.

        Args:
            encoder_output: output from the encoder
            encoder_output_length: lengths of the utterances in `encoder_output`
            prev_batched_state: previous batched decoding state
            multi_biasing_ids: optional tensor [Batch] with ids of multi-biasing models
        """
        assert self.cuda_graphs_mode is not None
        device = encoder_output.device

        # do not recalculate joint projection, project only once
        encoder_output = self.joint.project_encoder(encoder_output)
        current_batch_size = encoder_output.shape[0]
        current_max_time = encoder_output.shape[1]

        if torch.is_autocast_enabled():
            encoder_output = encoder_output.to(torch.get_autocast_gpu_dtype())
        else:
            # since autocast could be enabled outside and disallowed here,
            # we need to cast encoder output to dtype of params
            float_dtype = next(self.joint.parameters()).dtype
            encoder_output = encoder_output.to(float_dtype)

        # init or reinit graph
        if self.state is None or self.state.need_reinit(encoder_output):
            self._graph_reinitialize(encoder_output)

        # copy (projected) encoder output and lengths
        self.state.encoder_output_projected[:current_batch_size, :current_max_time, ...].copy_(encoder_output)
        self.state.encoder_output_length[: encoder_output_length.shape[0]].copy_(encoder_output_length)
        # set length to zero for elements outside the current batch
        self.state.encoder_output_length[current_batch_size:].fill_(0)
        if self.biasing_multi_model is not None:
            if multi_biasing_ids is None:
                multi_biasing_ids = torch.full(
                    [encoder_output_length.shape[0]], fill_value=-1, dtype=torch.long, device=device
                )
            self.state.multi_biasing_ids[:current_batch_size].copy_(multi_biasing_ids)
            self.state.multi_biasing_ids[current_batch_size:].fill_(-1)

        self._init_decoding_state(current_batch_size=current_batch_size, prev_batched_state=prev_batched_state)

        if self.cuda_graphs_mode is self.CudaGraphsMode.FULL_GRAPH:
            self.full_graph.replay()
        elif self.cuda_graphs_mode is self.CudaGraphsMode.NO_WHILE_LOOPS:
            self.separate_graphs.before_outer_loop.replay()
            while self.state.active_mask_any.item():
                self.separate_graphs.before_inner_loop.replay()
                while self.state.advance_mask_any.item():
                    self.separate_graphs.inner_loop_code.replay()
                self.separate_graphs.after_inner_loop.replay()
        elif self.cuda_graphs_mode is self.CudaGraphsMode.NO_GRAPHS:
            # this mode is only for testing purposes
            # manual loop instead of using graphs
            self._before_outer_loop()
            while self.state.active_mask_any.item():
                self._before_inner_loop_get_joint_output()
                while self.state.advance_mask_any.item():
                    self._inner_loop_step_find_next_non_blank()
                self._after_inner_loop_step()
        else:
            raise NotImplementedError(f"Unknown graph mode: {self.cuda_graphs_mode}")

        if prev_batched_state is not None:
            self._fix_timestamps_for_iterative_decoding(
                current_batch_size=current_batch_size, prev_batched_state=prev_batched_state
            )
        # NB: last labels can not exist (nothing decoded on this step).
        # return the last labels from the previous state in this case
        last_labels = self.state.batched_hyps.get_last_labels(pad_id=self._SOS)
        pad_batch_size = (
            self.state.batch_size - prev_batched_state.labels.shape[-1] if prev_batched_state is not None else 0
        )

        decoding_state = BatchedLabelLoopingState(
            predictor_states=self.decoder.clone_state(self.state.decoder_state),
            predictor_outputs=self.state.decoder_output.clone(),
            labels=(
                torch.where(
                    last_labels == self._SOS,
                    F.pad(prev_batched_state.labels, (0, pad_batch_size), value=self._SOS),
                    last_labels,
                )
                if prev_batched_state is not None
                else last_labels
            ),
            decoded_lengths=(
                self.state.encoder_output_length.clone()
                if prev_batched_state is None
                else self.state.encoder_output_length
                + F.pad(prev_batched_state.decoded_lengths, (0, pad_batch_size), value=0)
            ),
            fusion_states_list=([fusion_state.clone() for fusion_state in self.state.fusion_states_list]),
            time_jumps=self.state.time_indices - self.state.encoder_output_length,
        )

        # NB: return an independent copy of hyps/alignments/state
        # to avoid any manipulations with allocated memory outside the decoder
        return (
            self.state.batched_hyps.clone(),
            self.state.alignments.clone() if self.preserve_alignments else None,
            decoding_state,
        )

    @classmethod
    def _create_outer_while_loop_kernel(cls):
        """
        Creates a kernel that evaluates whether to enter the outer loop body (not all hypotheses are decoded).
        Condition: while(active_mask_any).
        """
        kernel_string = r"""\
        typedef __device_builtin__ unsigned long long cudaGraphConditionalHandle;

        extern "C" __device__ __cudart_builtin__ void cudaGraphSetConditional(cudaGraphConditionalHandle handle, unsigned int value);

        extern "C" __global__
        void outer_label_looping_conditional(cudaGraphConditionalHandle handle, const bool *active_mask_any)
        {
         cudaGraphSetConditional(handle, *active_mask_any);
        }
        """
        return run_nvrtc(kernel_string, b"outer_label_looping_conditional", cls.CUDA_PROGRAM_NAME)

    @classmethod
    def _create_inner_while_loop_kernel(cls):
        """
        Creates a kernel that evaluates whether to enter the inner loop body (not all non-blank labels found).
        Condition: while(advance_mask_any).
        """
        kernel_string = r"""\
        typedef __device_builtin__ unsigned long long cudaGraphConditionalHandle;

        extern "C" __device__ __cudart_builtin__ void cudaGraphSetConditional(cudaGraphConditionalHandle handle, unsigned int value);

        extern "C" __global__
        void inner_find_non_blank_conditional(cudaGraphConditionalHandle handle, const bool *advance_mask_any)
        {
         cudaGraphSetConditional(handle, *advance_mask_any);
        }
        """
        return run_nvrtc(kernel_string, b"inner_find_non_blank_conditional", cls.CUDA_PROGRAM_NAME)

    def _graph_reinitialize(
        self,
        encoder_output_projected: torch.Tensor,
    ):
        batch_size, max_time, encoder_dim = encoder_output_projected.shape

        self.state = LabelLoopingState(
            batch_size=batch_size,
            max_time=max(max_time, self.INITIAL_MAX_TIME),
            encoder_dim=encoder_dim,
            max_symbols=self.max_symbols,
            device=encoder_output_projected.device,
            float_dtype=encoder_output_projected.dtype,
            logits_dim=self.joint.num_classes_with_blank,
            preserve_alignments=self.preserve_alignments,
            preserve_frame_confidence=self.preserve_frame_confidence,
            include_duration_confidence=self.include_duration_confidence,
            include_duration=self.include_duration,
        )
        self.state.model_durations = self.durations.to(self.state.device, non_blocking=True)

        # init decoder state
        self.state.labels.fill_(self._SOS)
        decoder_output, new_state, *_ = self.decoder.predict(
            self.state.labels.unsqueeze(1),
            self.decoder.initialize_state(self.state.encoder_output_projected),
            add_sos=False,
            batch_size=self.state.batch_size,
        )
        self.state.decoder_state_after_sos = new_state
        self.state.decoder_state = self.decoder.initialize_state(encoder_output_projected)
        self.decoder.batch_replace_states_all(
            src_states=self.state.decoder_state_after_sos, dst_states=self.state.decoder_state
        )
        # to avoid recalculation of joint projection, store decoder output in state
        self.state.decoder_output_after_sos = self.joint.project_prednet(decoder_output)
        self.state.decoder_output = self.state.decoder_output_after_sos.clone()

        # init fusion models states and scores
        self.state.fusion_states_list = []
        self.state.fusion_states_candidates_list = []
        self.state.fusion_scores_list = []
        device = encoder_output_projected.device
        float_dtype = encoder_output_projected.dtype

        self._move_fusion_models_to_device(device=device)
        for fusion_model in self._all_fusion_models():
            vocab_size = fusion_model.vocab_size
            self.state.fusion_states_list.append(
                fusion_model.get_init_states(batch_size=self.state.batch_size, bos=True)
            )
            self.state.fusion_states_candidates_list.append(
                torch.zeros([batch_size, vocab_size], dtype=torch.long, device=device)
            )

            self.state.fusion_scores_list.append(
                torch.zeros([batch_size, vocab_size], dtype=float_dtype, device=device)
            )

        # warmup before graph compilation
        if self.cuda_graphs_mode is not self.CudaGraphsMode.NO_GRAPHS:
            self._warmup_for_cuda_graphs()

        if self.cuda_graphs_mode is self.CudaGraphsMode.FULL_GRAPH:
            try:
                self._full_graph_compile()
            except NeMoCUDAPythonException as e:
                if not self.cuda_graphs_allow_fallback:
                    raise RuntimeError("Full CUDA graph decoding failed. Mode is forced, raising exception") from e
                logging.warning(
                    f"Full CUDA graph compilation failed: {e}. "
                    "Falling back to native PyTorch CUDA graphs. Decoding will be slower."
                )
                self.cuda_graphs_mode = self.CudaGraphsMode.NO_WHILE_LOOPS
                self._partial_graphs_compile()
        elif self.cuda_graphs_mode is self.CudaGraphsMode.NO_WHILE_LOOPS:
            self._partial_graphs_compile()
        elif self.cuda_graphs_mode is self.CudaGraphsMode.NO_GRAPHS:
            # no graphs needed
            pass
        else:
            raise NotImplementedError

    def _warmup_for_cuda_graphs(self):
        """Warmup before compiling CUDA graphs"""
        is_ddp = torch.distributed.is_available() and torch.distributed.is_initialized()
        # 11 warmup steps required in DDP mode
        # see https://pytorch.org/docs/stable/notes/cuda.html#usage-with-distributeddataparallel
        num_runs = 11 if is_ddp else 3
        self.state.encoder_output_projected.fill_(0.0)
        self.state.encoder_output_length.fill_(1)
        s = torch.cuda.Stream()
        s.wait_stream(torch.cuda.current_stream())
        with torch.cuda.stream(s):
            for _ in range(num_runs):
                self._before_outer_loop()
                self._before_inner_loop_get_joint_output()
                self._inner_loop_step_find_next_non_blank()
                self._after_inner_loop_step()
        torch.cuda.current_stream().wait_stream(s)
        self.state.encoder_output_length.fill_(0)

    def _partial_graphs_compile(self):
        """Compile decoding by parts"""
        # Always create a new stream, because the per-thread default stream disallows stream capture to a graph.
        stream_for_graph = torch.cuda.Stream(self.state.device)
        stream_for_graph.wait_stream(torch.cuda.default_stream(self.state.device))
        self.separate_graphs = SeparateGraphsLabelLooping()
        with (
            torch.cuda.stream(stream_for_graph),
            torch.inference_mode(),
            torch.cuda.graph(
                self.separate_graphs.before_outer_loop, stream=stream_for_graph, capture_error_mode="thread_local"
            ),
        ):
            self._before_outer_loop()

        with (
            torch.cuda.stream(stream_for_graph),
            torch.inference_mode(),
            torch.cuda.graph(
                self.separate_graphs.before_inner_loop, stream=stream_for_graph, capture_error_mode="thread_local"
            ),
        ):
            self._before_inner_loop_get_joint_output()

        with (
            torch.cuda.stream(stream_for_graph),
            torch.inference_mode(),
            torch.cuda.graph(
                self.separate_graphs.inner_loop_code, stream=stream_for_graph, capture_error_mode="thread_local"
            ),
        ):
            self._inner_loop_step_find_next_non_blank()

        with (
            torch.cuda.stream(stream_for_graph),
            torch.inference_mode(),
            torch.cuda.graph(
                self.separate_graphs.after_inner_loop, stream=stream_for_graph, capture_error_mode="thread_local"
            ),
        ):
            self._after_inner_loop_step()

    def _full_graph_compile(self):
        """Compile full graph for decoding"""
        # Always create a new stream, because the per-thread default stream disallows stream capture to a graph.
        stream_for_graph = torch.cuda.Stream(self.state.device)
        stream_for_graph.wait_stream(torch.cuda.default_stream(self.state.device))
        self.full_graph = torch.cuda.CUDAGraph()
        with (
            torch.cuda.stream(stream_for_graph),
            torch.inference_mode(),
            torch.cuda.graph(self.full_graph, stream=stream_for_graph, capture_error_mode="thread_local"),
        ):
            self._before_outer_loop()

            capture_status, _, graph, _, _, _ = cu_call(
                cudart.cudaStreamGetCaptureInfo(torch.cuda.current_stream(device=self.state.device).cuda_stream)
            )
            assert capture_status == cudart.cudaStreamCaptureStatus.cudaStreamCaptureStatusActive

            # capture: while self.active_mask_any:
            (outer_loop_conditional_handle,) = cu_call(cudart.cudaGraphConditionalHandleCreate(graph, 0, 0))
            outer_loop_kernel = self._create_outer_while_loop_kernel()
            active_mask_any_ptr = np.array([self.state.active_mask_any.data_ptr()], dtype=np.uint64)
            outer_loop_args = np.array(
                [outer_loop_conditional_handle.getPtr(), active_mask_any_ptr.ctypes.data],
                dtype=np.uint64,
            )

            # loop while there are active utterances
            # while self.active_mask_any:
            with with_conditional_node(
                outer_loop_kernel, outer_loop_args, outer_loop_conditional_handle, device=self.state.device
            ):
                self._before_inner_loop_get_joint_output()
                # capture: while self.advance_mask_any.item():
                inner_while_loop_kernel = self._create_inner_while_loop_kernel()
                (inner_loop_conditional_handle,) = cu_call(cudart.cudaGraphConditionalHandleCreate(graph, 0, 0))
                advance_mask_any_ptr = np.array([self.state.advance_mask_any.data_ptr()], dtype=np.uint64)
                inner_loop_args = np.array(
                    [
                        inner_loop_conditional_handle.getPtr(),
                        advance_mask_any_ptr.ctypes.data,
                    ],
                    dtype=np.uint64,
                )
                # while self.advance_mask_any.item():
                with with_conditional_node(
                    inner_while_loop_kernel, inner_loop_args, inner_loop_conditional_handle, device=self.state.device
                ):
                    self._inner_loop_step_find_next_non_blank()
                self._after_inner_loop_step()

    def _init_decoding_state(
        self, current_batch_size: int, prev_batched_state: Optional[BatchedLabelLoopingState] = None
    ):
        # NB: we can speedup the case when prev_batched_state is None by using CUDA graphs
        if prev_batched_state is None:
            # last found labels - initially <SOS> (<blank>) symbol
            self.state.labels.fill_(self._SOS)
            self.decoder.batch_replace_states_all(
                src_states=self.state.decoder_state_after_sos, dst_states=self.state.decoder_state
            )
            self.state.decoder_output.copy_(self.state.decoder_output_after_sos)

            # init fusion models states
            for fusion_model_idx, fusion_model in enumerate(self._all_fusion_models()):
                self.state.fusion_states_list[fusion_model_idx].copy_(
                    fusion_model.get_init_states(batch_size=self.state.batch_size, bos=True)
                )

            self.state.time_indices.fill_(0)
        else:
            # labels
            self.state.labels[:current_batch_size].copy_(prev_batched_state.labels[:current_batch_size])
            # initial state
            self.decoder.batch_replace_states_all(
                src_states=prev_batched_state.predictor_states,
                dst_states=self.state.decoder_state,
                batch_size=current_batch_size,
            )
            self.state.decoder_output[:current_batch_size].copy_(
                prev_batched_state.predictor_outputs[:current_batch_size]
            )

            # init fusion models states
            for fusion_model_idx, fusion_model in enumerate(self._all_fusion_models()):
                self.state.fusion_states_list[fusion_model_idx][:current_batch_size].copy_(
                    prev_batched_state.fusion_states_list[fusion_model_idx][:current_batch_size]
                )

            self.state.time_indices[:current_batch_size].copy_(prev_batched_state.time_jumps[:current_batch_size])

    def _before_outer_loop(self):
        """Clear state and compute initial active mask"""
        self.state.batched_hyps.clear_()
        if self.state.alignments is not None:
            self.state.alignments.clear_()

        self.state.scores.fill_(0.0)

        # time indices
        self.state.time_indices_current_labels.copy_(self.state.time_indices)
        # safe time indices: guaranteed to be < encoder_output_length
        torch.sub(self.state.encoder_output_length, 1, out=self.state.last_timesteps)
        torch.minimum(self.state.time_indices, self.state.last_timesteps, out=self.state.safe_time_indices)

        # masks for utterances in batch
        torch.less(self.state.time_indices, self.state.encoder_output_length, out=self.state.active_mask)

        # same as: self.active_mask_any = active_mask.any()
        torch.any(self.state.active_mask, out=self.state.active_mask_any)
        self.state.durations.fill_(0)

    def _before_inner_loop_get_joint_output(self):
        """Get Joint output after decoder output, prepare inner loop to search for all next non-blank labels"""
        # stage 1: get joint output, iteratively seeking for non-blank labels
        # blank label in `labels` tensor means "end of hypothesis" (for this index)
        self.state.active_mask_prev.copy_(self.state.active_mask)
        logits = (
            self.joint.joint_after_projection(
                self.state.encoder_output_projected[self.state.batch_indices, self.state.safe_time_indices].unsqueeze(
                    1
                ),
                self.state.decoder_output,
            )
            .squeeze(1)
            .squeeze(1)
        )
        # same as: scores, labels = logits[:, : -self.state.model_durations.shape[0]].max(-1)
        torch.max(
            logits[:, : -self.state.model_durations.shape[0]], dim=-1, out=(self.state.scores, self.state.labels)
        )

        if self.has_fusion_models():
            fusion_scores_list, fusion_states_candidates_list = self.advance_fusion_models(
                fusion_states_list=self.state.fusion_states_list,
                multi_biasing_ids=self.state.multi_biasing_ids,
                float_dtype=self.state.float_dtype,
            )
            for fusion_model_idx in range(len(fusion_scores_list)):
                # get fusion scores/states
                self.state.fusion_states_candidates_list[fusion_model_idx].copy_(
                    fusion_states_candidates_list[fusion_model_idx]
                )
                self.state.fusion_scores_list[fusion_model_idx].copy_(fusion_scores_list[fusion_model_idx])
                # update logits with fusion scores
                logits[:, : -self.state.model_durations.shape[0] - 1] += fusion_scores_list[fusion_model_idx]
            # get labels (greedy) and scores from current logits, replace labels/scores with new
            scores_w_fusion, labels_w_fusion = logits[:, : -self.state.model_durations.shape[0] - 1].max(dim=-1)
            # preserve "blank" / "non-blank" category
            torch.where(
                self.state.labels == self._blank_index, self.state.labels, labels_w_fusion, out=self.state.labels
            )
            torch.where(
                self.state.labels == self._blank_index, self.state.scores, scores_w_fusion, out=self.state.scores
            )

        jump_durations_indices = logits[:, -self.state.model_durations.shape[0] :].argmax(dim=-1)
        self.state.durations.copy_(self.state.model_durations[jump_durations_indices])

        # search for non-blank labels using joint, advancing time indices for blank labels
        # checking max_symbols is not needed, since we already forced advancing time indices for such cases
        torch.eq(self.state.labels, self._blank_index, out=self.state.blank_mask)
        # blank_mask = self.labels == self._blank_index
        self.state.time_indices_current_labels.copy_(self.state.time_indices)
        # for blank labels force duration >= 1
        self.state.durations.masked_fill_(torch.logical_and(self.state.durations == 0, self.state.blank_mask), 1)

        if self.state.alignments is not None:
            self.state.alignments.add_results_masked_no_checks_(
                active_mask=self.state.active_mask,
                time_indices=self.state.time_indices_current_labels,
                logits=logits if self.preserve_alignments else None,
                labels=self.state.labels if self.preserve_alignments else None,
                confidence=self._get_frame_confidence(
                    logits=logits, num_durations=self.state.model_durations.shape[0]
                ),
            )

        # advance_mask is a mask for current batch for searching non-blank labels;
        # each element is True if non-blank symbol is not yet found AND we can increase the time index
        self.state.time_indices.add_(self.state.durations * self.state.active_mask)
        torch.minimum(self.state.time_indices, self.state.last_timesteps, out=self.state.safe_time_indices)
        torch.less(self.state.time_indices, self.state.encoder_output_length, out=self.state.active_mask)
        torch.logical_and(self.state.active_mask, self.state.blank_mask, out=self.state.advance_mask)

        # inner loop: find next non-blank labels (if exist)
        # same as: self.advance_mask_any = advance_mask.any()
        torch.any(self.state.advance_mask, out=self.state.advance_mask_any)

    def _inner_loop_step_find_next_non_blank(self):
        """Find next non-blank labels - one iteration"""
        # same as: time_indices_current_labels[advance_mask] = time_indices[advance_mask], but non-blocking
        # store current time indices to use further for storing the results
        torch.where(
            self.state.advance_mask,
            self.state.time_indices,
            self.state.time_indices_current_labels,
            out=self.state.time_indices_current_labels,
        )
        logits = (
            self.joint.joint_after_projection(
                self.state.encoder_output_projected[self.state.batch_indices, self.state.safe_time_indices].unsqueeze(
                    1
                ),
                self.state.decoder_output,
            )
            .squeeze(1)
            .squeeze(1)
        )
        # get labels (greedy) and scores from current logits, replace labels/scores with new
        # labels[advance_mask] are blank, and we are looking for non-blank labels
        more_scores, more_labels = logits[:, : -self.state.model_durations.shape[0]].max(-1)

        if self.has_fusion_models():
            for fusion_model_idx, fusion_scores in enumerate(self.state.fusion_scores_list):
                # update logits with fusion scores
                logits[:, : -self.state.model_durations.shape[0] - 1] += fusion_scores
            # # get labels (greedy) and scores from current logits, replace labels/scores with new
            more_scores_w_fusion, more_labels_w_fusion = logits[:, : -self.state.model_durations.shape[0] - 1].max(
                dim=-1
            )
            # preserve "blank" / "non-blank" category
            torch.where(more_labels == self._blank_index, more_labels, more_labels_w_fusion, out=more_labels)
            torch.where(more_labels == self._blank_index, more_scores, more_scores_w_fusion, out=more_scores)

        jump_durations_indices = logits[:, -self.state.model_durations.shape[0] :].argmax(dim=-1)
        more_durations = self.state.model_durations[jump_durations_indices]
        # same as: labels[advance_mask] = more_labels[advance_mask], but non-blocking
        torch.where(self.state.advance_mask, more_labels, self.state.labels, out=self.state.labels)
        # same as: scores[advance_mask] = more_scores[advance_mask], but non-blocking
        torch.where(self.state.advance_mask, more_scores, self.state.scores, out=self.state.scores)

        if self.state.alignments is not None:
            self.state.alignments.add_results_masked_no_checks_(
                active_mask=self.state.advance_mask,
                time_indices=self.state.time_indices_current_labels,
                logits=logits if self.preserve_alignments else None,
                labels=more_labels if self.preserve_alignments else None,
                confidence=self._get_frame_confidence(
                    logits=logits, num_durations=self.state.model_durations.shape[0]
                ),
            )

        # blank_mask = self.labels == self._blank_index
        torch.eq(self.state.labels, self._blank_index, out=self.state.blank_mask)
        # for blank labels force duration >= 1
        more_durations.masked_fill_(torch.logical_and(more_durations == 0, self.state.blank_mask), 1)
        # self.time_indices += self.blank_mask
        torch.where(
            self.state.advance_mask,
            self.state.time_indices + more_durations,
            self.state.time_indices,
            out=self.state.time_indices,
        )

        torch.where(self.state.advance_mask, more_durations, self.state.durations, out=self.state.durations)

        torch.minimum(self.state.time_indices, self.state.last_timesteps, out=self.state.safe_time_indices)
        torch.less(self.state.time_indices, self.state.encoder_output_length, out=self.state.active_mask)
        torch.logical_and(self.state.active_mask, self.state.blank_mask, out=self.state.advance_mask)
        torch.any(self.state.advance_mask, out=self.state.advance_mask_any)

    def _after_inner_loop_step(self):
        """After inner loop: store labels, query decoder/fusion models, force max symbols"""
        self._after_inner_loop_store_labels()
        self._after_inner_loop_select_fusion_states()
        self._after_inner_loop_get_decoder_output()
        self._after_inner_loop_force_max_symbols()

    def _after_inner_loop_store_labels(self):
        """Stage 3.1: Store hypotheses, update decoder state"""
        self.state.found_labels_mask.copy_(
            torch.logical_and(self.state.active_mask_prev, self.state.labels != self._blank_index)
        )
        self.state.batched_hyps.add_results_masked_no_checks_(
            active_mask=self.state.found_labels_mask,
            labels=self.state.labels,
            time_indices=self.state.time_indices_current_labels,
            scores=self.state.scores,
            token_durations=self.state.durations if self.include_duration else None,
        )

    def _after_inner_loop_select_fusion_states(self):
        """Stage 3.2: Select fusion states with new labels"""
        for fusion_model_idx, fusion_states_candidates in enumerate(self.state.fusion_states_candidates_list):
            # select necessary fusion states based on chosen labels
            torch.where(
                self.state.found_labels_mask,
                fusion_states_candidates[self.state.batch_indices, self.state.labels * self.state.found_labels_mask],
                self.state.fusion_states_list[fusion_model_idx],
                out=self.state.fusion_states_list[fusion_model_idx],
            )

    def _after_inner_loop_get_decoder_output(self):
        """Stage 3.3: Get decoder (prediction network) output using new labels"""
        decoder_output, new_state, *_ = self.decoder.predict(
            self.state.labels.unsqueeze(1), self.state.decoder_state, add_sos=False, batch_size=self.state.batch_size
        )
        self.decoder.batch_replace_states_mask(
            src_states=new_state, dst_states=self.state.decoder_state, mask=self.state.found_labels_mask
        )
        decoder_output_projected = self.joint.project_prednet(decoder_output)  # do not recalculate joint projection
        torch.where(
            self.state.found_labels_mask.unsqueeze(-1).unsqueeze(-1),
            decoder_output_projected,
            self.state.decoder_output,
            out=self.state.decoder_output,
        )

    def _after_inner_loop_force_max_symbols(self):
        """Stage 4: to avoid looping, go to next frame after max_symbols emission"""
        # if labels are non-blank (not end-of-utterance), check that last observed timestep with label:
        # if it is equal to the current time index, and number of observations is >= max_symbols, force blank
        force_blank_mask = torch.logical_and(
            self.state.active_mask,
            torch.logical_and(
                torch.logical_and(
                    self.state.labels != self._blank_index,
                    self.state.batched_hyps.last_timestamp_lasts >= self.max_symbols,
                ),
                self.state.batched_hyps.last_timestamp == self.state.time_indices,
            ),
        )
        self.state.time_indices.add_(force_blank_mask)  # emit blank => advance time indices
        # update safe_time_indices, non-blocking
        torch.minimum(self.state.time_indices, self.state.last_timesteps, out=self.state.safe_time_indices)
        # same as: active_mask = time_indices < encoder_output_length
        torch.less(self.state.time_indices, self.state.encoder_output_length, out=self.state.active_mask)
        torch.any(self.state.active_mask, out=self.state.active_mask_any)

    def _fix_timestamps_for_iterative_decoding(
        self, current_batch_size: int, prev_batched_state: BatchedLabelLoopingState
    ):
        """
        Fix timestamps: if we are in iterative decoding mode,
        we need to add the length of the previous batch to current timestamps
        """
        self.state.batched_hyps.timestamps[:current_batch_size] += prev_batched_state.decoded_lengths[
            :current_batch_size
        ].unsqueeze(1)
        if self.state.alignments is not None:
            self.state.alignments.timestamps[:current_batch_size] -= prev_batched_state.decoded_lengths[
                :current_batch_size
            ].unsqueeze(1)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
import pytest

from nemo.collections.llm import SpecterDataModule
from nemo.collections.llm.api import finetune
from nemo.collections.llm.recipes import e5_340m
from nemo.lightning import Trainer


class TestE5_340M:
    @pytest.fixture(scope="class")
    def recipe_module(self):
        return e5_340m

    def test_model(self, recipe_module):
        model_config = recipe_module.model()
        assert isinstance(model_config, run.Config)
        # Note: Actual model class assertions would depend on bert_embedding_model implementation
        # which isn't shown in the provided code

    def test_finetune_recipe(self, recipe_module):
        recipe = recipe_module.finetune_recipe()
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == finetune
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert isinstance(recipe.data, run.Config)
        assert recipe.data.__fn_or_cls__ == SpecterDataModule
        assert recipe.data.seq_length == 512
        assert recipe.data.global_batch_size == 32
        assert recipe.data.micro_batch_size == 4

    def test_finetune_recipe_with_custom_values(self, recipe_module):
        recipe = recipe_module.finetune_recipe(
            seq_length=1024,
            micro_batch_size=8,
            global_batch_size=64,
        )
        assert recipe.data.seq_length == 1024
        assert recipe.data.micro_batch_size == 8
        assert recipe.data.global_batch_size == 64

    @pytest.mark.parametrize("num_nodes,num_gpus_per_node", [(1, 8), (2, 4), (4, 2)])
    def test_finetune_recipe_with_different_configurations(self, recipe_module, num_nodes, num_gpus_per_node):
        recipe = recipe_module.finetune_recipe(num_nodes=num_nodes, num_gpus_per_node=num_gpus_per_node)
        assert recipe.trainer.num_nodes == num_nodes
        assert recipe.trainer.devices == num_gpus_per_node

    def test_finetune_recipe_with_peft(self, recipe_module):
        with pytest.raises(AssertionError, match="E5 only supports SFT."):
            recipe_module.finetune_recipe(peft_scheme='lora')

    def test_finetune_recipe_without_peft(self, recipe_module):
        recipe = recipe_module.finetune_recipe(peft_scheme=None)
        assert not hasattr(recipe, 'peft') or recipe.peft is None
